#!/usr/bin/env python

import sys
from typing import Optional

from dark.errors import ReadLengthsNotIdenticalError
from dark.reads import (
    Read,
    Reads,
    addFASTACommandLineOptions,
    parseFASTACommandLineOptions,
)


def printHeader(variableSites, args, baseOffset):
    n = len(variableSites)
    print(
        "%d site%s %svariable (threshold for homogeneity: %.3f)."
        % (
            n,
            " was" if n == 1 else "s were",
            "confirmed " if args.confirm else "",
            args.homogeneous,
        ),
        file=sys.stderr,
    )
    if args.reference:
        print(
            "Offsets adjusted by %d, relative to sequence %r."
            % (baseOffset, args.reference),
            file=sys.stderr,
        )


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description=(
            "Given FASTA on standard input, write reads with only the "
            "variable sites to standard output. If no sites are variable, "
            "nothing is written to standard output."
        )
    )

    parser.add_argument(
        "--printSites",
        action="store_true",
        default=False,
        help=(
            "Print the variable site locations to standard error along "
            "with the nucleotides found at the site, the homogeneity "
            "fraction of the site and the reference base (if a reference "
            "is given)."
        ),
    )

    parser.add_argument(
        "--sitesOnly",
        action="store_true",
        default=False,
        help=(
            "Only print the (comma separated) variable sites to standard "
            "output. This output can then be used directly as an argument "
            "for --keepSites to filter-fasta.py"
        ),
    )

    parser.add_argument(
        "--homogeneous",
        default=1.0,
        type=float,
        help=(
            "If the most-common nucleotide frequency at a site is at least "
            "this value, the site will be considered homogeneous."
        ),
    )

    parser.add_argument(
        "--confirm",
        action="store_true",
        default=False,
        help=(
            "Only keep sites where there is confirm variation (i.e., "
            "ambiguous sites that are compatible with there being no "
            "variation are not included)."
        ),
    )

    parser.add_argument(
        "--reference",
        help=(
            "A sequence id. Site offsets will be reported relative to this "
            "sequence. That means the number of gap (hyphen) chars at the "
            "start of this sequence will be counted and offsets will be "
            "incremented by that amount (and lower offsets in sequences "
            "that start before the specified sequence will be ignored)."
        ),
    )

    parser.add_argument(
        "--unknownAreAmbiguous",
        action="store_true",
        default=False,
        help=(
            "Any unknown character (e.g., a '-' gap or '?' unknown base) "
            "will be treated as being fully ambiguous (i.e., could be any "
            "of ACGT). Otherwise, all unknown characters are counted "
            "as '-' characters."
        ),
    )

    addFASTACommandLineOptions(parser)
    args = parser.parse_args()
    reads = Reads(list(parseFASTACommandLineOptions(args)))

    if not reads:
        sys.exit(0)

    reference: Optional[Read]

    if args.reference:
        for read in reads:
            if read.id == args.reference:
                break
        else:
            print(
                "Could not find --reference sequence %r." % args.reference,
                file=sys.stderr,
            )
            sys.exit(1)
        baseOffset = len(read.sequence) - len(read.sequence.lstrip("-"))
        reference = read
    else:
        baseOffset = 0
        reference = None

    try:
        variableSites = reads.variableSites(
            confirm=args.confirm,
            homogeneityLevel=args.homogeneous,
            unknownAreAmbiguous=args.unknownAreAmbiguous,
        )
    except ReadLengthsNotIdenticalError:
        print("Input sequences are not all the same length!", file=sys.stderr)
        sys.exit(2)

    if variableSites:
        toDelete = set()
        if args.printSites:
            for site, counts in variableSites.items():
                if site >= baseOffset:
                    ref = (" (ref %s)" % reference.sequence[site]) if reference else ""
                    print(
                        "%d: %s%s" % (site + 1 - baseOffset, counts, ref),
                        file=sys.stderr,
                    )
                else:
                    toDelete.add(site)

        for site in toDelete:
            del variableSites[site]

    if variableSites:
        if args.sitesOnly:
            print(
                ",".join(
                    map(lambda site: str(site + 1 - baseOffset), sorted(variableSites))
                )
            )
        else:
            saveAs = "fasta" if args.fasta else "fastq"
            reads.filter(keepSites=set(variableSites)).save(sys.stdout, saveAs)
            printHeader(variableSites, args, baseOffset)
    else:
        print(
            "No sites were %svariable (threshold for homogeneity: %.3f)."
            % ("confirmed " if args.confirm else "", args.homogeneous),
            file=sys.stderr,
        )



from typing import Iterator, Optional


def dimensionalIterator(
    dimensions: list[int], maxItems: int = -1, start: Optional[list[int]] = None
) -> Iterator[tuple[int, ...]]:
    """
    Given a list of n positive integers, return a generator that yields
    n-tuples of coordinates to 'fill' the dimensions. This is like an
    odometer in a car, but the dimensions do not each have to be 10.

    For example: dimensionalIterator((2, 3)) will yield in order
    (0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2). See the tests in
    test_dimension.py for many more examples.

    A dimension may also be given as '*', to provide a dimension that is
    never exhausted. For example, dimensionalIterator(('*', 2)) yields the
    infinite series (0, 0), (0, 1), (1, 0), (1, 1), (2, 0), (2, 1), ....

    maxItems can be used to limit the number of tuples yielded.
    """
    nDimensions = len(dimensions)
    if nDimensions == 0 or maxItems == 0:
        return
    if any(map(lambda x: x != "*" and x <= 0, dimensions)):
        raise ValueError("Dimensions not all positive! %r" % (dimensions,))

    if start:
        if len(start) != len(dimensions):
            raise ValueError(
                f"The start list is of length {len(start)}, but "
                f"the number of dimensions ({len(dimensions)}) differs."
            )
        odometer = list(start)
    else:
        odometer = [
            0,
        ] * nDimensions

    while maxItems != 0:
        yield tuple(odometer)
        maxItems -= 1
        wheel = nDimensions - 1
        while (
            dimensions[wheel] != "*"
            and odometer[wheel] == dimensions[wheel] - 1
            and wheel >= 0
        ):
            odometer[wheel] = 0
            wheel -= 1
        if wheel < 0:
            return
        odometer[wheel] += 1



#!/usr/bin/env python

"""
Extract aligned (i.e., padded) queries in FASTA format from a SAM/BAM file.
"""

import argparse
import sys

from dark.filter import (
    addFASTAFilteringCommandLineOptions,
    parseFASTAFilteringCommandLineOptions,
)
from dark.reads import Reads
from dark.sam import PaddedSAM, SAMFilter
from dark.utils import nucleotidesToStr

parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description="Produce aligned FASTA queries from a SAM/BAM file.",
)

parser.add_argument(
    "--rcSuffix",
    default="",
    help=(
        "A string to add to the end of query names that are reverse "
        "complemented. This is added before the /1, /2, etc., that are "
        "added for duplicated ids (if there are duplicates and "
        "--allowDuplicateIds is not used)"
    ),
)

parser.add_argument(
    "--rcNeeded",
    default=False,
    action="store_true",
    help=(
        "If given, queries that are flagged as matching when reverse "
        "complemented will be reverse complemented in the output. This "
        "must be used if the program that created the SAM/BAM input "
        "flags reversed matches but does not also store the reverse "
        "complemented query. The bwa program (mem and aln followed by "
        "samse) stores the queries reversed complemented if the match "
        "was, so this option is not needed for bwa. If in doubt, test the "
        "output of your matching program as this is very important!"
    ),
)

parser.add_argument(
    "--listReferenceInsertions",
    default=False,
    action="store_true",
    help=(
        "If given, information about reference sequence insertions will be "
        'printed to standard error. These correspond to "I" CIGAR '
        "operations that for the match would require inserting query bases "
        "into the reference. Because we cannot change the reference (in "
        "fact we typically do not have the reference in the SAM/BAM file), "
        "we cut the inserted bases out of the aligned query and save the "
        "information about what would have been inserted and where. That "
        "information is printed by this option. The output gives the "
        "0-based offset where the inserted base would be placed, followed "
        "by a list of the nucleotides that were suggested as being "
        "inserted and the number of times each nucleotide was suggested. "
        'So for example the output might contain "27: T:3, G:10" which '
        "indicates that 13 query (3 with T and 10 with G) matches would "
        "insert a nucleotide into the reference at offset 27."
    ),
)

SAMFilter.addFilteringOptions(parser)
addFASTAFilteringCommandLineOptions(parser)

args = parser.parse_args()
reads = parseFASTAFilteringCommandLineOptions(args, Reads())
samFilter = SAMFilter.parseFilteringOptions(args, filterRead=reads.filterRead)
paddedSAM = PaddedSAM(samFilter)

for read in paddedSAM.queries(rcSuffix=args.rcSuffix, rcNeeded=args.rcNeeded):
    print(read.toString("fasta"), end="")

if args.listReferenceInsertions:
    if paddedSAM.referenceInsertions:
        print(
            "(0-based) insertions into the reference:\n%s"
            % nucleotidesToStr(paddedSAM.referenceInsertions, "  "),
            file=sys.stderr,
        )
    else:
        print("No matches required an insertion into the reference.", file=sys.stderr)



#!/usr/bin/env python

import argparse
import multiprocessing
import sys
from math import log10

from dark.aligners import MAFFT_DEFAULT_ARGS, NEEDLE_DEFAULT_ARGS, align
from dark.dna import AMBIGUOUS, compareDNAReads, matchToString
from dark.reads import Reads, addFASTACommandLineOptions, parseFASTACommandLineOptions
from dark.utils import parseRangeExpression

MAFFT_ALGORITHMS_URL = (
    "https://mafft.cbrc.jp/alignment/software/algorithms/algorithms.html"
)


def getArgs() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Compare two DNA sequences.",
    )

    parser.add_argument(
        "--index1",
        type=int,
        default=1,
        help="The (1-based) index in the input of the first sequence.",
    )

    parser.add_argument(
        "--index2",
        type=int,
        default=2,
        help="The (1-based) index in the input of the second sequence.",
    )

    parser.add_argument(
        "--align",
        action="store_true",
        help=(
            "If given, use mafft (the default) or needle (according to the "
            "algorithm selected by --aligner) to align the two sequences."
        ),
    )

    parser.add_argument(
        "--aligner",
        default="mafft",
        choices=("edlib", "mafft", "needle"),
        help="The alignment algorithm to use.",
    )

    parser.add_argument(
        "--alignerOptions",
        help=(
            "Optional arguments to pass to the alignment algorithm. If the "
            f"aligner is 'mafft', the default options are {MAFFT_DEFAULT_ARGS!r}. "
            f"If 'needle', the default is {NEEDLE_DEFAULT_ARGS!r}. Do not try to set "
            "the number of threads here; use the --threads argument instead. If you "
            f"are using mafft, see {MAFFT_ALGORITHMS_URL} for some possible option "
            "combinations."
        ),
    )

    parser.add_argument(
        "--threads",
        type=int,
        default=multiprocessing.cpu_count(),
        help=(
            "The number of threads to use when running the aligner (if --align "
            "is used and the alignment algorithm can make use of multiple "
            "threads (mafft can, needle cannot))."
        ),
    )

    parser.add_argument(
        "--alignmentFile", help="The file to save the alignment to (implies --align)."
    )

    parser.add_argument(
        "--strict",
        action="store_true",
        help="If given, do not allow ambiguous nucleotide symbols to match.",
    )

    parser.add_argument(
        "--quiet",
        dest="verbose",
        action="store_false",
        help="Do not print information about aligning.",
    )

    parser.add_argument(
        "--noGapLocations",
        dest="includeGapLocations",
        action="store_false",
        help="Do not indicate the (1-based) locations of sequence gaps.",
    )

    parser.add_argument(
        "--noNoCoverageLocations",
        dest="includeCoverageLocations",
        action="store_false",
        help="Do not indicate the (1-based) locations of no coverage.",
    )

    parser.add_argument(
        "--sites",
        help=(
            "Specify (1-based) sequence sites to keep. All other sites will "
            "be ignored. The sites must be given in the form e.g., 24,100-200,260."
        ),
    )

    parser.add_argument(
        "--showDiffs",
        action="store_true",
        help="Print (1-based) sites where the sequence nucleotides differ.",
    )

    parser.add_argument(
        "--showAmbiguous",
        action="store_true",
        help=(
            "Print (1-based) sites where either sequence has an ambiguous "
            "nucleotide code."
        ),
    )

    parser.add_argument(
        "--includeAmbiguousMatches",
        action="store_true",
        help=(
            "Print (1-based) sites of ambiguous matches. The output gives the "
            "site, the base from the first sequence, then the base from the "
            "second sequence."
        ),
    )

    parser.add_argument(
        "--includeNonGapMismatches",
        action="store_true",
        help=(
            "Print (1-based) sites of mismatches that do not involve a gap. The "
            "output gives the site, the base from the first sequence, then the "
            "base from the second sequence."
        ),
    )

    parser.add_argument(
        "--includeDifferenceCounts",
        action="store_true",
        help=(
            "Include the counts (i.e., number of occurrences) of all differences. "
            "Note that mismatches involving areas of no coverage or where one read "
            "is shorter than another do not count as differences."
        ),
    )

    parser.add_argument(
        "--includeDifferenceLocations",
        action="store_true",
        help=(
            "Include the locations (and counts) (i.e., number of occurrences) "
            "of all differences. Note that mismatches involving areas of no coverage "
            "or where one read is shorter than another do not count as differences."
        ),
    )

    parser.add_argument(
        "--gapChars",
        default="-",
        metavar="CHARS",
        help=(
            "The sequence characters that should be considered to be gaps. "
            "These characters will be ignored in computing sequence lengths "
            "and identity fractions."
        ),
    )

    parser.add_argument(
        "--noCoverageChars",
        metavar="CHARS",
        help=(
            "The sequence characters that indicate lack of coverage. "
            "These characters will be ignored in identity fractions."
        ),
    )

    addFASTACommandLineOptions(parser)

    return parser.parse_args()


def main() -> None:
    args = getArgs()

    keepSequences = set([args.index1 - 1, args.index2 - 1])

    reads = list(parseFASTACommandLineOptions(args).filter(keepSequences=keepSequences))

    if len(reads) == 1:
        if len(keepSequences) == 1:
            # This is odd but not necessarily an error. For some reason, they're asking
            # for a comparison of a sequence with itself.
            if args.verbose:
                print(
                    "Warning: you are comparing a sequence with itself.",
                    file=sys.stderr,
                )

            reads = Reads([reads[0], reads[0]])
        else:
            sys.exit("Could not find both requested sequence indices. Exiting.")
    elif len(reads) != 2:
        sys.exit("Could not find both requested sequence indices. Exiting.")

    if args.alignmentFile:
        args.align = True

    if args.align:
        len1, len2 = map(len, reads)
        if len1 == len2:
            print("Pre-alignment, sequence lengths were identical: %d" % len1)
        else:
            print(
                "Pre-alignment, sequence lengths: %d, %d (difference %d)"
                % (len1, len2, abs(len1 - len2))
            )

        print("  Gaps:")
        print("    Id: %s %d" % (reads[0].id, reads[0].sequence.count("-")))
        print("    Id: %s %d" % (reads[1].id, reads[1].sequence.count("-")))

        reads = align(
            reads, args.aligner, args.alignerOptions, args.verbose, args.threads
        )

        if args.alignmentFile:
            assert reads.save(args.alignmentFile) == 2

    offsets = (
        parseRangeExpression(args.sites, convertToZeroBased=True)
        if args.sites
        else None
    )

    read1, read2 = reads
    len1, len2 = map(len, reads)
    identicalLengths = len1 == len2

    # Sanity check.
    if args.align:
        assert identicalLengths

    match = compareDNAReads(
        read1,
        read2,
        matchAmbiguous=(not args.strict),
        offsets=offsets,
        gapChars=args.gapChars,
        noCoverageChars=args.noCoverageChars,
    )

    x = "Post-alignment, sequence" if args.align else "Sequence"
    if identicalLengths:
        print("%s lengths are identical: %s" % (x, len1))
    else:
        print("%s lengths: %d, %d (difference %d)" % (x, len1, len2, abs(len1 - len2)))

    print(
        matchToString(
            match,
            read1,
            read2,
            matchAmbiguous=(not args.strict),
            offsets=offsets,
            includeGapLocations=args.includeGapLocations,
            includeNoCoverageLocations=args.includeCoverageLocations,
            includeAmbiguousMatches=args.includeAmbiguousMatches,
            includeNonGapMismatches=args.includeNonGapMismatches,
            includeDifferenceLocations=args.includeDifferenceLocations,
            includeDifferenceCounts=args.includeDifferenceCounts,
        )
    )

    if args.showDiffs:
        # Print all sites where the sequences differ.
        width = int(log10(max(len1, len2))) + 1
        headerPrinted = False
        for site, (a, b) in enumerate(zip(read1.sequence, read2.sequence), start=1):
            if a != b:
                if not headerPrinted:
                    print("Differences (site, %s, %s):" % (read1.id, read2.id))
                    headerPrinted = True
                print("  %*d %s %s" % (width, site, a, b))

        if not headerPrinted:
            print("No sequence differences found.")

    if args.showAmbiguous:
        width = int(log10(max(len1, len2))) + 1
        headerPrinted = False
        for site, (a, b) in enumerate(zip(read1.sequence, read2.sequence), start=1):
            if len(AMBIGUOUS.get(a, "")) > 1 or len(AMBIGUOUS.get(b, "")) > 1:
                if not headerPrinted:
                    print("Ambiguities (site, %s, %s):" % (read1.id, read2.id))
                    headerPrinted = True
                print("  %*d %s %s" % (width, site, a, b))

        if not headerPrinted:
            print("No sequence ambiguities found.")


if __name__ == "__main__":
    main()



#!/usr/bin/env python

from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description=(
            "Given FASTA on stdin, write the ids, sequences, and "
            "quality strings on a single TAB-separated line to stdout."
        )
    )

    parser.add_argument(
        "--separator",
        default="\t",
        help=(
            "The character string to separate ids from "
            "sequences and quality strings (if any)"
        ),
    )

    parser.add_argument(
        "--removeIds", action="store_true", help="Do not print sequence ids"
    )

    addFASTACommandLineOptions(parser)
    args = parser.parse_args()
    sep = args.separator
    reads = parseFASTACommandLineOptions(args)

    # Duplicate code a little so as not to repeatedly do two tests per read.
    if args.fastq:
        if args.removeIds:
            for read in reads:
                print(read.sequence + sep + read.quality)
        else:
            for read in reads:
                print(read.id + sep + read.sequence + sep + read.quality)
    else:
        if args.removeIds:
            for read in reads:
                print(read.sequence)
        else:
            for read in reads:
                print(read.id + sep + read.sequence)



from unittest import TestCase

from dark.graphics import plotAAProperties
from dark.reads import AARead


class TestPlotAAProperties(TestCase):
    """
    Tests for the plotAAProperties function in graphics.py
    """

    # Note that plotAAProperties relies on dark.aa.propertiesForSequence
    # which is tested thoroughly in test_aa.py

    def testNoProperties(self):
        """
        plotAAProperties must run correctly when called with no properties.
        """
        read = AARead("id", "AI")
        self.assertEqual({}, plotAAProperties(read, [], showFigure=False))

    def testOneProperty(self):
        """
        plotAAProperties must run correctly when called with one property.
        """
        read = AARead("id", "AI")
        self.assertEqual(
            {
                "composition": [-1.0, -1.0],
            },
            plotAAProperties(read, ["composition"], showFigure=False),
        )

    def testTwoProperties(self):
        """
        plotAAProperties must run correctly when called with two properties.
        """
        read = AARead("id", "AI")
        self.assertEqual(
            {
                "composition": [-1.0, -1.0],
                "hydropathy": [0.4, 1.0],
            },
            plotAAProperties(read, ["composition", "hydropathy"], showFigure=False),
        )



from random import choice, choices
from unittest import TestCase

from dark.aligners import (
    EDLIB_AMBIGUOUS,
    canonicalizeAllGaps,
    edlibAlign,
    removeUnnecessaryGaps,
)
from dark.dna import AMBIGUOUS
from dark.reads import DNARead, Reads


class TestEdlibAlign(TestCase):
    """
    Test the edlibAlign function.
    """

    def testAmbiguousCodes(self):
        """
        Test that the edlib ambiguous code tuples are correct.
        """
        # Check a few random tuples (see ../dark/dna.py for more).
        self.assertIn(("A", "M"), EDLIB_AMBIGUOUS)
        self.assertIn(("G", "R"), EDLIB_AMBIGUOUS)
        self.assertNotIn(("R", "G"), EDLIB_AMBIGUOUS)
        self.assertIn(("A", "N"), EDLIB_AMBIGUOUS)

    def testNumberOfAmbiguousCodes(self):
        """
        Test the number of edlib ambiguous code tuples. The 28 comes from
        summing the lengths of all the ambiguous sets of length > 1
        (see ../dark/dna.py).
        """
        expected = sum(
            len(ambiguities)
            for ambiguities in AMBIGUOUS.values()
            if len(ambiguities) > 1
        )

        self.assertEqual(expected, 28)
        self.assertEqual(expected, len(EDLIB_AMBIGUOUS))

    def testEmtpyStrings(self):
        """
        Aligning two empty reads results in an Exception.
        """
        in1 = DNARead("id1", "")
        in2 = DNARead("id2", "")

        error = (
            r"^The object alignResult contains an empty CIGAR string\. "
            r"Users must run align\(\) with task='path'\. Please check "
            r"the input alignResult\.$"
        )
        self.assertRaisesRegex(Exception, error, edlibAlign, Reads([in1, in2]))

    def testTooManySequences(self):
        """
        Passing more than two sequences must result in a ValueError if
        onlyTwoSequences is True.
        """
        in1 = DNARead("id1", "")
        in2 = DNARead("id2", "")
        in3 = DNARead("id3", "")

        error = r"^Passed 1 unexpected extra sequences\.$"
        self.assertRaisesRegex(Exception, error, edlibAlign, Reads([in1, in2, in3]))

    def testFirstReadAlreadyHasAGap(self):
        """
        If the first read has a gap character we should get a ValueError.
        """
        in1 = DNARead("id1", "-A")
        in2 = DNARead("id2", "AA")

        error = r"^Sequence 'id1' contains one or more gap characters '-'\.$"
        self.assertRaisesRegex(Exception, error, edlibAlign, Reads([in1, in2]))

    def testSecondReadAlreadyHasAGap(self):
        """
        If the second read has a gap character we should get a ValueError.
        """
        in1 = DNARead("id1", "AACCT")
        in2 = DNARead("id2", "AA--T")

        error = r"^Sequence 'id2' contains one or more gap characters '-'\.$"
        self.assertRaisesRegex(Exception, error, edlibAlign, Reads([in1, in2]))

    def testIdenticalStringsOfLengthOne(self):
        """
        Aligning identical reads of length one must produce the expected
        result.
        """
        in1 = DNARead("id1", "A")
        in2 = DNARead("id2", "A")

        out1, out2 = list(edlibAlign(Reads([in1, in2])))

        self.assertEqual(in1, out1)
        self.assertEqual(in2, out2)

    def testThreeIdenticalStringsOfLengthOne(self):
        """
        Aligning identical reads of length one must produce the expected
        result, including if we pass three sequences and onlyTwoSequences
        is False.
        """
        in1 = DNARead("id1", "A")
        in2 = DNARead("id2", "A")
        in3 = DNARead("id3", "A")

        out1, out2 = list(edlibAlign(Reads([in1, in2, in3]), onlyTwoSequences=False))

        self.assertEqual(in1, out1)
        self.assertEqual(in2, out2)

    def testNonIdenticalStringsOfLengthOne(self):
        """
        Aligning non-identical reads of length one must produce the expected
        result.
        """
        in1 = DNARead("id1", "A")
        in2 = DNARead("id2", "G")

        out1, out2 = list(edlibAlign(Reads([in1, in2])))

        self.assertEqual(in1, out1)
        self.assertEqual(in2, out2)

    def testIdenticalStringsOfLengthTwo(self):
        """
        Aligning identical reads of length two must produce the expected
        result.
        """
        in1 = DNARead("id1", "AG")
        in2 = DNARead("id2", "AG")

        out1, out2 = list(edlibAlign(Reads([in1, in2])))

        self.assertEqual(in1, out1)
        self.assertEqual(in2, out2)

    def testNonIdenticalStringsOfLengthTwo(self):
        """
        Aligning non-identical reads of length two must produce the expected
        result.
        """
        in1 = DNARead("id1", "AT")
        in2 = DNARead("id2", "GC")

        out1, out2 = list(edlibAlign(Reads([in1, in2])))

        self.assertEqual(in1, out1)
        self.assertEqual(in2, out2)

    def testGappedPrefix(self):
        """
        Aligning a read that is a suffix of another must result in gaps being
        placed at the start of the shorter string.
        """
        in1 = DNARead("id1", "ATGCCGTTCA")
        in2 = DNARead("id2", "GCCGTTCA")

        out1, out2 = list(edlibAlign(Reads([in1, in2])))

        self.assertEqual(in1, out1)
        self.assertEqual("--GCCGTTCA", out2.sequence)

    def testLongerGappedPrefix(self):
        """
        Aligning a read that is a suffix of another must result in gaps being
        placed at the expected places at the start of the shorter string.
        """
        in1 = DNARead("id1", "ATGCCGTTCA")
        in2 = DNARead("id2", "CGTTCA")

        out1, out2 = list(edlibAlign(Reads([in1, in2])))

        self.assertEqual(in1, out1)
        self.assertEqual("----CGTTCA", out2.sequence)

    def testGappedSuffix(self):
        """
        Aligning a read that is a prefix of another must result in gaps being
        placed at the end of the shorter string.
        """
        in1 = DNARead("id1", "ATGCCGTTCA")
        in2 = DNARead("id2", "ATGCCGTT")

        out1, out2 = list(edlibAlign(Reads([in1, in2])))

        self.assertEqual(in1, out1)
        self.assertEqual("ATGCCGTT--", out2.sequence)

    def testLongerGappedSuffix(self):
        """
        Aligning a read that is a prefix of another must result in gaps being
        placed at the expected places at the end of the shorter string.
        """
        in1 = DNARead("id1", "ATGCCGTTCA")
        in2 = DNARead("id2", "ATGCC")

        out1, out2 = list(edlibAlign(Reads([in1, in2])))

        self.assertEqual(in1, out1)
        self.assertEqual("ATGCC-----", out2.sequence)

    def testLongerGappedSuffixWithGap(self):
        """
        Aligning a read that is a prefix of another must result in gaps being
        placed at the expected places at the end of the shorter string.
        """
        in1 = DNARead("id1", "ATGCCGTTCA")
        in2 = DNARead("id2", "ATGCCTT")

        out1, out2 = list(edlibAlign(Reads([in1, in2])))

        self.assertEqual("ATGCCGTTCA", out1.sequence)
        self.assertEqual("ATGCC-TT--", out2.sequence)

    def testGapsAtBothEnds(self):
        """
        Aligning two reads that have a common prefix/suffix must result
        in gaps at the start of one and the end of the other.
        """
        in1 = DNARead("id1", "XXXATGCCGTTCA")
        in2 = DNARead("id2", "ATGCCGTTCAYYY")

        out1, out2 = list(edlibAlign(Reads([in1, in2])))

        self.assertEqual("XXXATGCCGTTCA---", out1.sequence)
        self.assertEqual("---ATGCCGTTCAYYY", out2.sequence)

    def testGapsAtBothEndsAndOneInTheMiddle(self):
        """
        Aligning two reads that have a common prefix/suffix and one deletion
        must result in gaps at the start of one and the end of the other,
        and one in the middle.
        """
        in1 = DNARead("id1", "XXXATGCCTTCA")
        in2 = DNARead("id2", "ATGCCGTTCAYYY")

        out1, out2 = list(edlibAlign(Reads([in1, in2])))

        self.assertEqual("XXXATGCC-TTCA---", out1.sequence)
        self.assertEqual("---ATGCCGTTCAYYY", out2.sequence)

    def testGapsAtBothEndsAndTwoInTheMiddle(self):
        """
        Aligning two reads that have a common prefix/suffix and two deletions
        must result in gaps at the start of one and the end of the other,
        and two in the middle.
        """
        in1 = DNARead("id1", "XXXATCCTTCA")
        in2 = DNARead("id2", "ATGCCGTTCAYYY")

        out1, out2 = list(edlibAlign(Reads([in1, in2])))

        self.assertEqual("XXXAT-CC-TTCA---", out1.sequence)
        self.assertEqual("---ATGCCGTTCAYYY", out2.sequence)

    def testOneMiddleGapRemovalNotMinimized(self):
        """
        Test removal of one gap from the middle of each sequence when
        unnecessary gaps are not removed.
        """
        in1 = DNARead("id1", "TAACGCAGA")
        in2 = DNARead("id2", "TAAGTCAGA")

        out1, out2 = list(edlibAlign(Reads([in1, in2]), minimizeGaps=False))

        self.assertEqual("TAACG-CAGA", out1.sequence)
        self.assertEqual("TAA-GTCAGA", out2.sequence)

    def testOneMiddleGapRemovalMinimized(self):
        """
        Test removal of one gap from the middle of each sequence when
        unnecessary gaps are removed.
        """
        in1 = DNARead("id1", "TAACGCAGA")
        in2 = DNARead("id2", "TAAGTCAGA")

        out1, out2 = list(edlibAlign(Reads([in1, in2])))

        self.assertEqual(in1, out1)
        self.assertEqual(in2, out2)

    def testTwoMiddleGapsRemovalNotMinimized(self):
        """
        Test removal of two gaps from the middle of each sequence when the
        post-processing step to remove unnecessary gaps is not done.
        """
        in1 = DNARead("id1", "TAACGCAGAxTAACGCAGA")
        in2 = DNARead("id2", "TAAGTCAGAxTAAGTCAGA")

        out1, out2 = list(edlibAlign(Reads([in1, in2]), minimizeGaps=False))

        self.assertEqual("TAACG-CAGAxTAACG-CAGA", out1.sequence)
        self.assertEqual("TAA-GTCAGAxTAA-GTCAGA", out2.sequence)

    def testTwoMiddleGapsRemovalMinimized(self):
        """
        Test removal of two gaps from the middle of each sequence when
        unnecessary gaps are removed.
        """
        in1 = DNARead("id1", "TAACGCAGAxTAACGCAGA")
        in2 = DNARead("id2", "TAAGTCAGAxTAAGTCAGA")

        out1, out2 = list(edlibAlign(Reads([in1, in2])))

        self.assertEqual(in1, out1)
        self.assertEqual(in2, out2)

    def test30000Nucleotides(self):
        """
        Aligning two reads of length 30K nucleotides must work (and run
        quickly).
        """
        in1 = DNARead("id1", "A" * 30_000)
        in2 = DNARead("id2", "A" * 30_000)

        out1, out2 = list(edlibAlign(Reads([in1, in2])))

        self.assertEqual(in1, out1)
        self.assertEqual(in2, out2)

    def testNoWildcardMatch(self):
        """
        If wildcards are not used...
        """
        in1 = DNARead("id1", "ACTG")
        in2 = DNARead("id2", "ATG")

        out1, out2 = list(edlibAlign(Reads([in1, in2])))

        self.assertEqual("ACTG", out1.sequence)
        self.assertEqual("A-TG", out2.sequence)

    def test500SubstitutionsIn30000Nucleotides(self):
        """
        Test that when 500 random substitutions (though not next to each
        other) are made to a random 30K sequence, that the alignment is as
        expected (i.e., that the sequences are perfectly aligned despite the
        mismatches).
        """
        n = 30_000
        nSubs = 500
        seq1 = choices("ACGT", k=n)
        seq2 = seq1[:]

        # Pick some substitution sites, but not next to one another
        # (otherwise there is a chance gaps are introduced).
        subOffsets = set()
        while len(subOffsets) < nSubs:
            offset = choice(range(n))
            if not (subOffsets & {offset - 1, offset, offset + 1}):
                subOffsets.add(offset)

        # Make some changes in seq2.
        for offset in subOffsets:
            b4 = seq2[offset]
            others = list(set("ACGT") - {seq2[offset]})
            self.assertNotIn(b4, others)
            seq2[offset] = choice(others)

        seq1 = "".join(seq1)
        seq2 = "".join(seq2)

        in1 = DNARead("id1", seq1)
        in2 = DNARead("id2", seq2)

        out1, out2 = list(edlibAlign(Reads([in1, in2])))

        self.assertEqual(len(in1), len(out1))
        self.assertEqual(in1, out1)
        self.assertEqual(seq2, out2.sequence)

    def test500DeletionsIn30000Nucleotides(self):
        """
        Test that when 500 random deletions (though not next to each other) are
        made to a random 30K sequence, that the alignment has gaps at the
        expected sites.
        """
        n = 30_000
        nDels = 500

        # Make a first sequence that doesn't have repeated nucelotides.
        # This makes sure edlib doesn't have an adjacent choice about where
        # to put a gap in a string of identical nucelotides.
        seq1 = []
        while len(seq1) < n:
            nt = choice("ACGT")
            if not seq1 or seq1[-1] != nt:
                seq1.append(nt)

        # Pick some deletion sites, but not next to one another (otherwise
        # it is too hard to know what gaps to test for below).
        delOffsets = set()
        while len(delOffsets) < nDels:
            offset = choice(range(n))
            if not (delOffsets & {offset - 1, offset, offset + 1}):
                delOffsets.add(offset)

        # seq2 will be a copy of seq1, but sans the deleted sites.
        seq2 = []
        for offset in range(n):
            if offset not in delOffsets:
                seq2.append(seq1[offset])

        seq1 = "".join(seq1)
        seq2 = "".join(seq2)

        in1 = DNARead("id1", seq1)
        in2 = DNARead("id2", seq2)

        out1, out2 = list(edlibAlign(Reads([in1, in2])))

        # Check that all the sites that were deleted in seq2 now have a gap.
        for offset in range(n):
            if offset in delOffsets:
                # There is (maybe) a small chance the assertion below will
                # fail, in which case the following code can be uncommented
                # to see what is going on. An info line will be printed,
                # with a '*' at offsets where there should be a gap but
                # isn't and a '|' at gap offsets. This is followed by the
                # two aligned sequences, with '-' characters. I'm leaving
                # this code here for now, until we are very sure it doesn't
                # fail due to some pathological selection of substitution
                # sites and nucelotides. If that does happen, reducing
                # nDels or broadening the exclusion zone around each
                # deletion (to more than +/- 1 nt) should greatly reduce
                # the probability of this test failing.
                #
                # if out2.sequence[offset] != '-':
                #     for a in range(n):
                #         print('*' if a == offset else (
                #             '|' if a in delOffsets else ' '), end='')
                #     print()
                #     print(out1.sequence)
                #     print(out2.sequence)
                self.assertEqual("-", out2.sequence[offset])


class TestRemoveUnnecessaryGaps(TestCase):
    """
    Test the removeUnnecessaryGaps function.
    """

    def testEmptyStrings(self):
        """
        No gaps can be removed from two empty strings.
        """
        self.assertEqual(removeUnnecessaryGaps("", ""), ("", ""))

    def testStringsWithNoGaps(self):
        """
        No gaps can be removed from two strings with no gaps.
        """
        self.assertEqual(removeUnnecessaryGaps("AG", "AT"), ("AG", "AT"))

    def testOneGapOneIdenticalChar(self):
        """
        Identical strings of length two with one gap have the gap removed.
        """
        self.assertEqual(removeUnnecessaryGaps("A-", "-A"), ("A", "A"))
        self.assertEqual(removeUnnecessaryGaps("-A", "A-"), ("A", "A"))

    def testOneGapOneNonIdenticalChar(self):
        """
        Non-identical strings of length two with one gap have the gap removed.
        """
        self.assertEqual(removeUnnecessaryGaps("-A", "B-"), ("A", "B"))
        self.assertEqual(removeUnnecessaryGaps("A-", "-B"), ("A", "B"))

    def testFirstGapInFirstString(self):
        """
        A gap can be removed from two strings when the first string is the
        first to have a gap.
        """
        self.assertEqual(
            removeUnnecessaryGaps("TAA-GTCAGA", "TAACG-CAGA"),
            ("TAAGTCAGA", "TAACGCAGA"),
        )

    def testFirstGapInSecondString(self):
        """
        A gap can be removed from two strings when the second string is the
        first to have a gap.
        """
        self.assertEqual(
            removeUnnecessaryGaps("TAACG-CAGA", "TAA-GTCAGA"),
            ("TAACGCAGA", "TAAGTCAGA"),
        )

    def testConsecutiveGaps(self):
        """
        Two consecutive gaps can be removed.
        """
        self.assertEqual(
            removeUnnecessaryGaps("TAA--GTCAGA", "TAACG--CAGA"),
            ("TAAGTCAGA", "TAACGCAGA"),
        )


class TestCanonicalizeAllGaps(TestCase):
    """
    Test the canonicalizeAllGaps function.
    """

    def testEmptyStrings(self):
        """
        No gaps can be canonicalized in two empty strings.
        """
        self.assertEqual(
            canonicalizeAllGaps(
                "",
                "",
            ),
            (
                "",
                "",
            ),
        )

    def testStringsWithNoGaps(self):
        """
        No gaps can be canonicalized from two strings with no gaps.
        """
        self.assertEqual(
            canonicalizeAllGaps(
                "AG",
                "AT",
            ),
            (
                "AG",
                "AT",
            ),
        )

    def testLengthTwo(self):
        """
        The simplest fixable case with sequences of length two. Here the
        gap in the second sequence can be shifted left because the A at
        the start of the second sequence also matches the second A in the
        first sequence.
        """
        self.assertEqual(
            canonicalizeAllGaps(
                "AA",
                "A-",
            ),
            (
                "AA",
                "-A",
            ),
        )

    def testLengthTwoTwice(self):
        """
        The simplest fixable case with sequences of length two. Here the
        gap in the second sequence can be shifted left because the A at
        the start of the second sequence also matches the second A in the
        first sequence. We pass sequences in which this can be done twice.
        """
        self.assertEqual(
            canonicalizeAllGaps(
                "AAGCC",
                "A-GC-",
            ),
            (
                "AAGCC",
                "-AG-C",
            ),
        )

    def testLengthThree(self):
        """
        A simple case with sequences of length three. Here the
        gap in the second sequence can be shifted left because the A at
        the start of the second sequence also matches the final A in the
        first sequence.
        """
        self.assertEqual(
            canonicalizeAllGaps(
                "AAA",
                "A--",
            ),
            (
                "AAA",
                "--A",
            ),
        )

    def testLengthFour(self):
        """
        A simple case with sequences of length four. Here the
        gap in the second sequence can be shifted left because the A at
        the start of the second sequence also matches the final A in the
        first sequence.
        """
        self.assertEqual(
            canonicalizeAllGaps(
                "AAAT",
                "AT--",
            ),
            (
                "AAAT",
                "--AT",
            ),
        )

    def testOneIsolated(self):
        """
        An isolated nucelotide should be moved to the right.
        """
        self.assertEqual(
            canonicalizeAllGaps(
                "ACTCAGACGTC",
                "ACT--G---TC",
            ),
            (
                "ACTCAGACGTC",
                "ACT-----GTC",
            ),
        )

    def testOneIsolatedCanonicalizeFromTheRight(self):
        """
        An isolated nucelotide should be moved to the right. But this is more
        tricky because the "CT" in the second sequence could be moved right,
        producing "A--CTG---TC". We need the "G" to be moved to the right. That
        creates a gap that the "ACT", "CT", and "T" prefixes cannot then be
        moved to the right of.
        """
        self.assertEqual(
            canonicalizeAllGaps(
                "ACTCTGACGTC",
                "ACT--G---TC",
            ),
            (
                "ACTCTGACGTC",
                "ACT-----GTC",
            ),
        )

    def testLengthTwoMoveThree(self):
        """
        A gap of length two should be able to be moved three places to the left.
        """
        self.assertEqual(
            canonicalizeAllGaps(
                "TAAGGGGATCAGA",
                "TAAGGA--GCAGA",
            ),
            (
                "TAAGGGGATCAGA",
                "TAA--GGAGCAGA",
            ),
        )

    def testLengthThreeMoveTwo(self):
        """
        A gap of length three should be able to be moved two places to the left.
        """
        self.assertEqual(
            canonicalizeAllGaps(
                "ACTCGACGTC",
                "ACTCG---TC",
            ),
            (
                "ACTCGACGTC",
                "ACT---CGTC",
            ),
        )

    def testLongExample(self):
        """
        A more complicated example, in which all gaps are merged.
        """
        self.assertEqual(
            canonicalizeAllGaps(
                "ATTAGGTTTATACCTTCCCAGGTAACAAACCAACTTTCGATCTCTTGTAGATCTGTTCTCTAAACGAAC",
                "A---G----AT-C-T-----G-T-----------T--C--TCT-----A-A----------A--CGAAC",
            ),
            (
                "ATTAGGTTTATACCTTCCCAGGTAACAAACCAACTTTCGATCTCTTGTAGATCTGTTCTCTAAACGAAC",
                "------------------------------------------------AGATCTGTTCTCTAAACGAAC",
            ),
        )

    def testLongExampleOneNtLeftOnLeftMultiple(self):
        """
        A more complicated example, in which all gaps are merged and one nucleotide
        is left on the left-hand side. Done with many concatenated copies, just to
        prove we can.
        """
        seq_1 = "ATTAGGTTTATACCTTCCCAGGTAACAAACCAACTTTCGATCTCTTGTAGATCTGTTCTCTAAACGAAC"
        seq_2 = "C---G----AT-C-T-----G-T-----------T--C--TCT-----A-A----------A--CGAAC"
        canon = "C------------------------------------------------GATCTGTTCTCTAAACGAAC"

        for n in range(10):
            self.assertEqual(
                canonicalizeAllGaps(seq_1 * n, seq_2 * n),
                (seq_1 * n, canon * n),
            )



#!/usr/bin/env python

import argparse
import csv
import sys
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor
from contextlib import redirect_stdout
from itertools import repeat
from time import time
from typing import Any, Collection

import numpy as np
import pysam

from dark.fasta import FastaReads
from dark.sam import samfile, samReferenceLengths
from dark.utils import baseCountsToStr, entropy2, gmt, pct

BASES = "ACGT"
TRANSITIONS = "AG GA CT TC".split()
TRANSVERSIONS = "AT TA AC CA GT TG GC CG".split()


def makeParser():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            "Print SAM/BAM per-site nucleotide composition and coverage statistics. "
            "Note that no output is produced for sites with no read coverage."
        ),
    )

    parser.add_argument(
        "samfile",
        help="The SAM/BAM file to read.",
    )

    parser.add_argument(
        "--referenceId",
        metavar="ID",
        help=(
            "A reference sequence id. If not given and the SAM/BAM file has just one "
            "reference, that one will be used. Otherwise, a reference id must be given."
        ),
    )

    parser.add_argument(
        "--fasta",
        help=(
            "Optionally give a FASTA file in which the reference can be found. "
            "If you do not provide a reference sequence, mutations will be counted "
            "as differences from the most-common nucleotide at each genome site."
        ),
    )

    parser.add_argument(
        "--diffsFrom",
        choices=("reference", "commonest"),
        default="reference",
        help=(
            "The base from which to assess differences. If 'reference', differences "
            "from the reference base (at each genome site) will be treated as "
            "mutations. If 'commonest', differences from the most common base will be "
            "treated as mutations (with ties broken in favour of the reference base, "
            "if a reference is given). It is important that you understand the "
            "difference between these two and choose the right option."
        ),
    )

    parser.add_argument(
        "--window",
        "-n",
        type=int,
        default=250,
        help=(
            "The number of genome sites to have each subprocess handle. Large values "
            "will have a major performance impact when --maxDepth is also high (e.g., "
            ">10,000). This seems to be due to an unknown pysam/htslib slowdown (based "
            "on samtools v1.21, as of 2025-01-31)."
        ),
    )

    parser.add_argument(
        "--workers",
        type=int,
        help=(
            "The number of subprocesses to run in parallel. If not specified, the "
            "process pool will typically use all available cores, but see "
            "https://docs.python.org/3/library/concurrent.futures.html"
            "#processpoolexecutor for details."
        ),
    )

    parser.add_argument(
        "--noOffsets",
        action="store_false",
        dest="printOffsets",
        help="Do not print per-site details of base counts.",
    )

    parser.add_argument(
        "--noStats",
        action="store_false",
        dest="printStats",
        help="Do not print final average, standard deviation etc. statistics.",
    )

    parser.add_argument(
        "--statsFile",
        help=(
            "The file to print overall statistics to. If not specified, standard "
            "output will be used."
        ),
    )

    parser.add_argument(
        "--verbosity",
        type=int,
        default=0,
        help=(
            "Print progress information (to standard error). 0 = don't print progress "
            "information. 1 = print information after each genome region is processed. "
            "2 = also print information during the processing of each region."
        ),
    )

    parser.add_argument(
        "--minDepth",
        type=int,
        help=(
            "The minimum read depth to report on. Sites with a lower depth will "
            "not be reported."
        ),
    )

    parser.add_argument(
        "--maxDepth",
        type=int,
        default=1e6,
        help="The maximum read depth to request from pysam for each site.",
    )

    parser.add_argument(
        "--minSiteMutationRate",
        type=float,
        help=(
            "The minimum site mutation rate to report on. Sites with a lower mutation "
            "rate will not be reported. The mutation rate is the fraction of bases at "
            "the site that are not the reference or commonest base (depending on "
            "the --diffsFrom option)."
        ),
    )

    parser.add_argument(
        "--maxSiteMutationRate",
        type=float,
        help=(
            "The maximum site mutation rate to report on. Sites with a higher mutation "
            "rate will not be reported. The mutation rate is the fraction of bases at "
            "the site that are not the reference or commonest base (depending on "
            "the --diffsFrom option)."
        ),
    )

    parser.add_argument(
        "--minSiteMutationCount",
        type=int,
        default=0,
        help=(
            "The minimum number of mutant nucleotides that must be present at a site "
            "in order for it to be reported. E.g., if this were set to 4, there would "
            "need to be at least four nucleotides (of any mixture of bases) that were "
            "not the consensus or reference (depending on the --diffsFrom setting)."
        ),
    )

    parser.add_argument(
        "--minEntropy",
        type=float,
        default=0.0,
        help=(
            "The minimum site entropy to report on. Sites with lower entropy "
            "will not be reported."
        ),
    )

    parser.add_argument(
        "--maxEntropy",
        type=float,
        default=2.0,
        help=(
            "The maximum site entropy to report on. Sites with higher entropy "
            "will not be reported."
        ),
    )

    parser.add_argument(
        "--noFilter",
        action="store_false",
        dest="filter",
        help="This behaviour is now the default and this option is ignored.",
    )

    parser.add_argument(
        "--tsv",
        action="store_true",
        help=(
            "Produce TSV output (use --sep , if you want CSV). Note that overall "
            "SAM/BAM file statistics will be printed at the bottom of the output "
            "unless you also specify --noStats or --statsFile. Columns will hopefully "
            "be self-explanatory, but a few clarifying comments will probably be "
            "helpful. 1) The 'Base' column indicates the nucleotide base that was used "
            "to count mutations. The base used will depend on the value of --diffsFrom. "  # noqa: E501
            "See the help output for that option. 2) The 'Reference' column will have "
            "an empty value if no reference sequence was provided. 3) The 'Site' column "  # noqa: E501
            "gives 1-based indexes into the reference genome."
        ),
    )

    parser.add_argument(
        "--sep",
        default="\t",
        help="The separator to use when --tsv is used. Ignored otherwise.",
    )

    parser.add_argument(
        "--minBases",
        type=int,
        choices=(1, 2, 3, 4),
        default=1,
        help=(
            "The minimum number of different bases that must occur at a site for "
            "the site to be reported. E.g., if this were set to 3, sites are "
            "only counted if at most one nucleotide is not present. The default "
            "value of 1 will do no filtering because all (covered) sites will have "
            "at least one nucleotide present. Higher values of this parameter can "
            "be used to add a diversity filter, by only counting sites at which "
            "there is greater nucleotide diversity."
        ),
    )

    parser.add_argument(
        "--maxIdenticalReadIdsPerSite",
        type=int,
        default=2,
        help=(
            "The number of reads with identical IDs that are permitted at each site. "
            "This should be 2 for paired-end reads (Illumina sequencing uses the "
            "identical read ID for both members of the pair) because the read pair "
            "may overlap at some sites. If you have unpaired reads, you could set this "
            "to 1. A factor that needs to be considered is whether the mapping "
            "algorithm was told to only produce the best match for each read (or read "
            "pair). If so, setting this to 1 or 2 (as just described) makes sense "
            "because it will catch inadvertent errors. But if reads are allowed to "
            "match multiple times, there could be many matches of the same read in "
            "the same reference section. In that case, you can use a value of -1 to "
            "disable the checking."
        ),
    )

    parser.add_argument(
        "--minBaseQuality",
        type=int,
        default=13,
        help=(
            "The minimum nucleotide base quality. Note that samtools depth defaults "
            "to 13."
        ),
    )

    parser.add_argument(
        "--minMapQuality",
        type=int,
        default=0,
        help="The minimum mapping quality (SAM MAPQ field).",
    )

    return parser


def getReferenceId(
    filename: str, references: Collection[str], referenceId: str | None
) -> str:
    """
    Get the reference id.  This is either the reference id from the command line
    (if given and if present in the SAM/BAM file) or the single reference from the
    SAM/BAM file (if it contains just one reference).

    @param filename: The SAM/BAM filename.
    @param references: The names of all references in the SAM/BAM file.
    @param referenceId: The id of the wanted reference, if provided on the command line.
    @return: The reference id.
    """
    if referenceId:
        if referenceId not in references:
            raise ValueError(
                f"Reference id {referenceId} not found in {filename!r} "
                f"Known references are {', '.join(sorted(references))}."
            )
    else:
        if len(references) == 1:
            referenceId = list(references)[0]
        else:
            raise ValueError(
                f"Multiple references found in {filename!r}: "
                f"{', '.join(sorted(references))}. Use --referenceId "
                "to specify the one you want."
            )

    return referenceId


def getReferenceSeq(filename: str, referenceId: str) -> str:
    """
    Look for the reference id in the FASTA file and return its sequence if found.
    If no reference is found, raise a ValueError.

    @param filename: The FASTA filename.
    @param referenceId: The ID of the wanted reference.
    @return: The reference nucleotide sequence.
    """
    reads = list(FastaReads(filename))

    # First look for matches with the full id+description sequence.
    for read in reads:
        if read.id == referenceId:
            return read.sequence

    # Then look for matches with the first (space separated) part of the full
    # id+description sequence. We do this second so as to allow the full id+description
    # to be attempted first and only if that doesn't work do we allow for a
    # less-complete match.
    for read in reads:
        if read.id.split()[0] == referenceId:
            return read.sequence

    raise ValueError(
        f"Reference id {referenceId!r} was not found in reference FASTA "
        f"file {filename!r}."
    )


def analyzeColumn(
    column: pysam.PileupColumn,
    start: int,
    window: int,
    minDepth: int | None,
    minSiteMutationRate: float | None,
    maxSiteMutationRate: float | None,
    minSiteMutationCount: int,
    minEntropy: float,
    maxEntropy: float,
    minBases: int,
    maxIdenticalReadIdsPerSite: int,
    verbosity: int,
    referenceSeq: str | None,
    diffsFrom: str,
) -> (
    tuple[
        int,
        set[str],
        float,
        float,
        dict[str, int],
        int,
        str | None,
        str,
        dict[str, int],
    ]
    | None
):
    """
    Do an analysis of the reads in a column. Return a tuple of results
    unless the column should be skipped, in which case return None.

    @param start: The start offset of the genome window we process.
    @param window: The size of the genome section we should process.
    @param miDepth: The minimum read depth to report on. Sites with lower depth will
        be ignored.
    @param minSiteMutationRate: The minimum mutation rate (i.e., fraction) required at
        sites in order that they are reported. Sites with lower mutation fractions will
        be ignored.
    @param maxSiteMutationRate: The maximum mutation rate (i.e., fraction) allowed at
        sites in order that they are reported. Sites with higher mutation fractions will
        be ignored.
    @param minSiteMutationCount: The minimum number of mutant nucleotides that must be
        present at a site in order for it to be reported. E.g., if this were set to 4,
        there would need to be at least four nucleotides (of any mixture of bases) that
        were not the consensus or reference (depending on the --diffsFrom setting).
    @param minEntropy: The minimum entropy to allow.
    @param maxEntropy: The maximum entropy to allow.
    @param minBases: The minimum number of different bases that must occur at a site for
        the site to be reported. See the parser argument above for more details (or run
        this script with --help).
    @param maxIdenticalReadIdsPerSite: The number of reads with identical IDs that are
        permitted at each site. Use -1 to turn off the checking. See the parser argument
        above for more details (or run this script with --help).
    @param verbosity: The verbosity level.
    @param referenceSeq: The genome of the reference (if known).
    @param diffsFrom: The base from which to assess differences. If 'reference',
        differences from the reference base (at each genome site) will be treated as
        mutations. If 'commonest', differences from the most common base will be
        treated as mutations (with ties broken in favour of the reference base,
        if a reference is given).
    @return: A tuple containing the following variables (types are above in the
        typehint), whose names are hopefully self-explanatory:

            read ids
            mutation rate
            entropy
            mutation pair counts
            reference offset
            reference base
            from base
            bases
    """
    refOffset = column.reference_pos
    assert start <= refOffset < start + window

    # readIds will have read ids as keys, with values that are lists containing:
    #
    #   the number of times the read has been seen in this column (site).
    #   the current highest quality for the base for that read at this column (site).
    #   the base with the best quality.
    #
    # This information is collected because both reads of a read pair can
    # overlap the same site.  We want to choose the base that has the highest
    # quality and to detect if there is somehow more than a single read pair
    # (with the same read id) matching a site (this could happen if a read
    # mapping algorithm maps the same read pair more than once and the
    # regions of the two matches overlaps).
    readIds: dict[str, list[Any]] = {}

    for read in column.pileups:
        # assert isinstance(read, pysam.PileupRead)
        if read.query_position is None:
            assert read.is_del or read.is_refskip
        else:
            # assert isinstance(read.alignment, pysam.AlignedSegment)
            readId = read.alignment.query_name
            assert readId
            assert read.alignment.query_sequence is not None  # Typing check.
            base = read.alignment.query_sequence[read.query_position]

            if base not in BASES:
                # In practice this should never happen because the FASTQ we analyze
                # always only has ACGT bases. Allowing non-ACGT bases through here
                # results in a KeyError below when we count base frequencies using a
                # dict.fromkeys(BASES, 0). So issue a warning and continue.
                print(
                    f"WARNING: Site {refOffset + 1} is matched by a read "
                    f"(id {readId!r}) containing a non-ACGT base ({base}). "
                    "This is very unusual, as FASTQ data should normally only "
                    "contain sequences composed of unambiguous ACGT nucleotide calls.",
                    file=sys.stderr,
                )
                continue

            assert read.alignment.query_qualities is not None  # Typing check.
            quality: str = read.alignment.query_qualities[read.query_position]
            try:
                count, previousQuality, previousBase = readIds[readId]
            except KeyError:
                # We've not seen this read at this site before. Remember the
                # details so we can compare against them later (in the 'else'
                # just below) if the read pair also overlaps this site.
                readIds[readId] = [1, quality, base]
            else:
                count += 1
                if (
                    maxIdenticalReadIdsPerSite > 0
                    and count > maxIdenticalReadIdsPerSite
                ):
                    raise ValueError(
                        f"Read id {readId} occurs (at least) {count} times for site "
                        f"{refOffset + 1}. The maximum allowed repeats of the same read "  # noqa: E501
                        f"ID at a site is {maxIdenticalReadIdsPerSite}."
                    )
                # This is either the read pair or another match of the same
                # read to this site.
                readIds[readId][0] = count

                if base != previousBase:
                    if quality > previousQuality:  # Note that this is a string compare.
                        readIds[readId][1] = quality
                        readIds[readId][2] = base
                        if verbosity:
                            print(
                                f"Site {refOffset + 1}: replaced base "
                                f"({previousBase!r} with {base!r}) due to improved "
                                "quality in second read.",
                                file=sys.stderr,
                            )
                    elif quality == previousQuality:
                        # We should probably choose the consensus (or
                        # reference, if known) base here, if they are among
                        # the options. But we don't know what the consensus
                        # base is, yet. This may not be a problem as a later
                        # read (with the same ID) might have a base with a
                        # higher quality.
                        if verbosity > 1:
                            print(
                                f"Site {refOffset + 1}: has multiple bases "
                                f"({base!r} and {previousBase!r}) of the same quality. "
                                f"Ignoring {base!r}. Note that we may encounter "
                                "another read for this site with a higher quality, in "
                                "which case these equally-lower-quality base calls "
                                "will both be ignored in its favour.",
                                file=sys.stderr,
                            )

    # Note that if we have paired-end reads and they overlap the site, we count
    # them both because we are trying to figure out the depth and these are two
    # independent measurements of the same nucleotide. We could instead just
    # count them once (like samtools depth -s) but I don't think that's what
    # depth should really mean, since to me it's about confidence in the nt
    # found at a location and if we have an overlapping read pair, we should
    # count each read separately.
    nReads = sum(readInfo[0] for readInfo in readIds.values())

    if nReads == 0 or (minDepth is not None and nReads < minDepth):
        # nReads can still be zero at this point because all reads
        # matching this site have a read.query_position of None (due
        # to having deletion or reference skips).
        if verbosity > 1:
            print(
                f"Site {refOffset + 1}: skipped due to too few reads ({nReads:,}).",
                file=sys.stderr,
            )
        return

    bases = dict.fromkeys(BASES, 0)
    for count, _, base in readIds.values():
        bases[base] += count

    assert nReads == sum(bases.values())

    # Set an initial nonsense entropy value to keep type checking from
    # warning that the variable may be unbound.
    entropy = -1.0

    nt = []
    for base, count in bases.items():
        nt.extend([base] * count)

    entropy = entropy2(nt)

    if entropy < minEntropy:
        if verbosity > 1:
            print(
                f"Site {refOffset + 1}: skipped due to low entropy ({entropy:.5f}).",
                file=sys.stderr,
            )
        return

    if entropy > maxEntropy:
        if verbosity > 1:
            print(
                f"Site {refOffset + 1}: skipped due to high entropy ({entropy:.5f}).",
                file=sys.stderr,
            )
        return

    refBase = referenceSeq[refOffset] if referenceSeq else None

    if diffsFrom == "reference":
        assert refBase is not None
        fromBase = refBase
    else:
        # Find the base with the highest count. This is the 'from' base
        # (i.e., the originally existing base that was mutated from something
        # to something else). Ties are broken in favour of the reference base
        # (if there is no reference, the second value in the compared tuples
        # will always be False (because refBase will be None)). If there are
        # two equally frequent bases, neither of which is the reference (if
        # known), the tie is broken in favour of the alphabetically first
        # base.
        fromBase = max(
            (count, int(base == refBase), base) for (base, count) in bases.items()
        )[2]

    mutationCount = 0
    mutationPairCounts: dict[str, int] = defaultdict(int)

    try:
        mutationRate = 1.0 - bases[fromBase] / nReads
    except KeyError:
        # The reference might have a non-ACGT base. If so, we can't calculate a
        # mutation rate in a reliable/easy way, so we set it to 0.0. An
        # alternative would be to get the possible bases and to count the reads
        # which differ from all the possible bases as being mutations and the
        # rest not. But we still couldn't individually categorize them in
        # mutationPairCounts.
        mutationRate = 0.0
    else:
        mutationCount = sum(bases.values()) - bases[fromBase]
        for base in BASES:
            if base != fromBase and (count := bases[base]):
                mutationPairCounts[fromBase + base] += count

        if minSiteMutationRate is not None and mutationRate < minSiteMutationRate:
            if verbosity > 1:
                print(
                    f"Site {refOffset + 1}: skipped due to low mutation rate "
                    f"({mutationRate:.5f}).",
                    file=sys.stderr,
                )
            return

        if maxSiteMutationRate is not None and mutationRate > maxSiteMutationRate:
            if verbosity > 1:
                print(
                    f"Site {refOffset + 1}: skipped due to high mutation rate "
                    f"({mutationRate:.5f}).",
                    file=sys.stderr,
                )
            return

        for base in BASES:
            if base != fromBase and (count := bases[base]):
                mutationPairCounts[fromBase + base] += count
                mutationCount += count

        if minSiteMutationCount > 0 and mutationCount < minSiteMutationCount:
            if verbosity > 1:
                print(
                    f"Site {refOffset + 1}: skipped due to having insufficient "
                    f"nucleotide diversity ({mutationCount} different nucleotides "
                    "are present, but at least {minSiteMutationCount} are required).",
                    file=sys.stderr,
                )
            return

    # The number of bases at the site is the 'from' base, plus the bases it
    # has mutated to. Hence the + 1 in the following test.
    if len(mutationPairCounts) + 1 < minBases:
        if verbosity > 1:
            print(
                f"Site {refOffset + 1}: skipped due to having insufficient "
                f"nucleotide diversity ({len(mutationPairCounts) + 1} different "
                f"nucleotides are present, but at least {minBases} are required).",
                file=sys.stderr,
            )
        return

    return (
        nReads,
        set(readIds),
        mutationRate,
        entropy,
        mutationPairCounts,
        refOffset,
        refBase,
        fromBase,
        bases,
    )


def makeRow(
    refOffset: int,
    refBase: str | None,
    nReads: int,
    bases: dict[str, int],
    fromBase: str,
    mutationRate: float,
    mutationPairCounts: dict[str, int],
    entropy: float,
) -> list[int | str | float | None]:
    """
    Make a row of data (for a site) to be added to the rows returned by
    C{processWindowColumns}.

    @param refOffset: The 0-based offset into the reference.
    @param refBase: The reference base (if the reference is known).
    @param nReads: The number of reads for the column being examined.
    @param bases: The count of each nucleotide base at this site.
    @param fromBase: The base that is considered canonical at this site (i.e.,
        the one that mutations are measured from).
    @param mutationRate: The fraction of bases at the site that are mutations.
    @param mutationPairCounts: The count of each mutation pair at this site, where keys
        are the concatenated 'from' and 'to' bases involved in the mutation (e.g.,
        mutationPairCounts["CT"] holds the number of C -> T mutations).
    @param entropy: The entropy of the bases at this genome offset.
    """
    mutationCount = sum(mutationPairCounts.values())

    row = (
        [refOffset + 1, refBase, nReads]
        # Don't use bases.values() in the following, because we need bases in
        # ACGT order.
        + [bases[base] for base in BASES]
        + [
            fromBase,
            len(mutationPairCounts) + 1,
            mutationCount,
            round(mutationRate, 6),
            round(entropy, 6),
        ]
    )

    totalTransitions = 0
    for pair in TRANSITIONS:
        count = mutationPairCounts.get(pair, 0)
        row.append(count)
        totalTransitions += count

    totalTransversions = 0
    for pair in TRANSVERSIONS:
        count = mutationPairCounts.get(pair, 0)
        row.append(count)
        totalTransversions += count

    row.extend((totalTransitions, totalTransversions))

    return row


def processWindowColumns(
    filename: str,
    start: int,
    window: int,
    minDepth: int | None,
    maxDepth: int,
    minBaseQuality: int,
    minMapQuality: int,
    minSiteMutationRate: float | None,
    maxSiteMutationRate: float | None,
    minSiteMutationCount: int,
    minEntropy: float,
    maxEntropy: float,
    minBases: int,
    maxIdenticalReadIdsPerSite: int,
    collectOffsets: bool,
    verbosity: int,
    referenceId: str,
    referenceSeq: str | None,
    returnList: bool,
    startTime: float,
    diffsFrom: str,
) -> dict[str, Any]:
    """
    Process all columns in a subsection (window) of the genome.

    @param filename: The SAM/BAM file to read.
    @param start: The start offset of the genome window we process.
    @param window: The size of the genome section we should process.
    @param miDepth: The minimum read depth to report on. Sites with lower depth will
        be ignored.
    @param maxDepth: The maximum read depth to consider.
    @param minBaseQuality: The minimum base quality required for a base to be
        considered.
    @param minMapQuality: The minimum mapping quality required for a read to be
        considered.
    @param minSiteMutationRate: The minimum mutation rate (i.e., fraction) required at
        sites in order that they are reported. Sites with lower mutation fractions will
        be ignored.
    @param maxSiteMutationRate: The maximum mutation rate (i.e., fraction) allowed at
        sites in order that they are reported. Sites with higher mutation fractions will
        be ignored.
    @param minSiteMutationCount: The minimum number of mutant nucleotides that must be
        present at a site in order for it to be reported. E.g., if this were set to 4,
        there would need to be at least four nucleotides (of any mixture of bases) that
        were not the consensus or reference (depending on the --diffsFrom setting).
    @param minBases: The minimum number of different bases that must occur at a site for
        the site to be reported. See the parser argument above for more details (or run
        this script with --help).
    @param maxIdenticalReadIdsPerSite: The number of reads with identical IDs that are
        permitted at each site. Use -1 to turn off the checking. See the parser argument
        above for more details (or run this script with --help).
    @param collectOffsets: Whether to collect genome offset information.
    @param verbosity: The verbosity level.
    @param referenceId: The name of the reference sequence.
    @param referenceSeq: The genome of the reference (if known).
    @param returnList: If True, the rows we return will be printed (by our caller) as
        CSV, so we return lists. Else, we return a string for each row of output.
    @param startTime: The time overall processing was launched.
    @param diffsFrom: The base from which to assess differences. If 'reference',
        differences from the reference base (at each genome site) will be treated as
        mutations. If 'commonest', differences from the most common base will be
        treated as mutations (with ties broken in favour of the reference base,
        if a reference is given).
    """
    startTime = time()
    allReadIds = set()
    rows = []
    depths = []
    cumulativeReadCount = mutationCount = 0
    baseCounts = dict.fromkeys(BASES, 0)
    totalMutationPairCounts = defaultdict(int)

    with samfile(filename) as sam:
        for column in sam.pileup(
            start=start,
            end=start + window,
            reference=referenceId,
            truncate=True,
            max_depth=maxDepth,
            # We could add a --samtools option to use samtools-compatible
            # behavior. But see the warning about also needing to pass a
            # fastafile argument in that case at
            # https://pysam.readthedocs.io/en/latest/
            # api.html#pysam.AlignmentFile.pileup
            #
            # stepper="samtools",
            min_base_quality=minBaseQuality,  # Equivalent to samtools depth -Q
            min_mapping_quality=minMapQuality,  # Equivalent to samtools depth -q
            ignore_orphans=False,
            ignore_overlaps=False,
        ):
            assert isinstance(column, pysam.PileupColumn)  # Typing check.
            columnStartTime = time()

            columnInfo = analyzeColumn(
                column,
                start,
                window,
                minDepth,
                minSiteMutationRate,
                maxSiteMutationRate,
                minSiteMutationCount,
                minEntropy,
                maxEntropy,
                minBases,
                maxIdenticalReadIdsPerSite,
                verbosity,
                referenceSeq,
                diffsFrom,
            )

            if columnInfo is None:
                # This column should be skipped. It didn't have enough diversity or
                # reads, or failed some other check.
                continue

            (
                nReads,
                columnReadIds,
                mutationRate,
                entropy,
                mutationPairCounts,
                refOffset,
                refBase,
                fromBase,
                columnBaseCounts,
            ) = columnInfo

            if diffsFrom == "commonest":
                assert columnBaseCounts[fromBase] >= max(columnBaseCounts.values())

            # We cannot assert 'column.get_num_aligned() != nReads' here because
            # some of the reads may have been filtered out due to being
            # insertions or reference skips (deletions). See 'assert read.is_del
            # or read.is_refskip' in analyze_column above.

            depths.append(nReads)
            allReadIds.update(columnReadIds)
            mutationCount += sum(mutationPairCounts.values())

            for base, count in columnBaseCounts.items():
                baseCounts[base] += count

            for pair, count in mutationPairCounts.items():
                totalMutationPairCounts[pair] += count

            if collectOffsets:
                if returnList:
                    rows.append(
                        makeRow(
                            refOffset,
                            refBase,
                            nReads,
                            columnBaseCounts,
                            fromBase,
                            mutationRate,
                            mutationPairCounts,
                            entropy,
                        )
                    )
                else:
                    out = (
                        f"{refOffset + 1} {nReads} {baseCountsToStr(columnBaseCounts)}"
                    )
                    if referenceSeq:
                        out += f" Ref:{refBase}"
                    rows.append(out)

            if verbosity > 1:
                cumulativeReadCount += nReads
                stopTime = time()
                elapsed = stopTime - columnStartTime
                totalElapsed = stopTime - startTime

                print(
                    f"Site {refOffset + 1}: {nReads:,} reads processed "
                    f"at {nReads / elapsed / 1e6:.2f} M read/s. "
                    f"Total bases: {sum(baseCounts.values()):,} "
                    f"Overall: {cumulativeReadCount / totalElapsed / 1e6:.2f} M read/s",
                    file=sys.stderr,
                )

    return {
        "depths": depths,
        "baseCounts": baseCounts,
        "mutationPairCounts": totalMutationPairCounts,
        "readIds": allReadIds,
        "rows": rows,
        "startSite": start,
        "startTime": startTime,
        "stopTime": time(),
        "mutationCount": mutationCount,
    }


def printStats(
    statsFile,
    samfile,
    minDepth,
    maxDepth,
    diffsFrom,
    minBases,
    maxIdenticalReadIdsPerSite,
    minBaseQuality,
    minMapQuality,
    referenceSeq,
    referenceId,
    referenceLen,
    startGMT,
    elapsed,
    readIds,
    depths,
    minSiteMutationRate,
    maxSiteMutationRate,
    minSiteMutationCount,
    minEntropy,
    maxEntropy,
    totalBaseCounts,
    mutationPairCounts,
) -> None:
    """
    Print a summary of the run. See the docstrings elsewhere if you want to know what
    all these parameters mean :-)
    """
    with open(statsFile, "w") if statsFile else sys.stdout as out, redirect_stdout(out):
        print("Inputs:")
        print("  SAM/BAM file:", samfile)
        print("  Reference:")
        print("    ID:", referenceId)
        print("    Length:", referenceLen)
        print("    Sequence given:", "yes" if referenceSeq else "no")
        print("Settings:")
        print("  Max identical read IDs per site:", maxIdenticalReadIdsPerSite)
        print("  Depth:")
        print("    Min depth required:", minDepth)
        print("    Max depth considered:", maxDepth)
        print("  Quality:")
        print("    Min base quality:", minBaseQuality)
        print("    Min mapping quality:", minMapQuality)
        print("  Mutation (by site):")
        print("    Counts are relative to:", diffsFrom)
        print("    Min mutation rate:", minSiteMutationRate)
        print("    Max mutation rate:", maxSiteMutationRate)
        print(f"    Min entropy: {minEntropy:.5f}")
        print(f"    Max entropy: {maxEntropy:.5f}")
        print(f"    Min non-{diffsFrom} nucleotides:", minSiteMutationCount)
        print("    Min different nucleotides:", minBases)
        print("Result:")
        print("  Timing:")
        print("    Started at:", startGMT)
        print("    Stopped at:", gmt())
        print(f"    Elapsed: {elapsed} seconds")
        print(
            f"  Number of reads covering sites of interest: {sum(depths)} "
            f"({len(readIds)} unique read ids)"
        )
        print("  Reference genome sites included:", pct(len(depths), referenceLen))
        print("  Coverage depth:")
        print("    Min:", min(depths, default=0))
        print("    Max:", max(depths, default=0))

        if depths:
            print(f"    Mean: {np.mean(depths):.3f}")
            print(f"    Median: {np.median(depths):.3f}")
            print(f"    SD: {np.std(depths):.3f}")

        print("  Bases:")
        totalBases = 0
        for base, count in totalBaseCounts.items():
            print(f"    {base}: {count}")
            totalBases += count
        print(f"    TOTAL: {totalBases}")

        mutationCount = sum(mutationPairCounts.values())
        overallMutationRate = mutationCount / totalBases if totalBases else 0.0

        print("  Mutations:")
        print(f"    Overall rate: {overallMutationRate:.6f}")

        totalTransitions = totalTransversions = 0
        print("    Sorted counts (* = transversion):")
        countWidth = None

        def key(pair):
            return mutationPairCounts[pair], pair

        for pair in sorted(mutationPairCounts, key=key, reverse=True):
            count = mutationPairCounts[pair]
            if countWidth is None:
                countWidth = len(str(count))
            if pair in TRANSITIONS:
                totalTransitions += count
                what = ""
            else:
                assert pair in TRANSVERSIONS
                totalTransversions += count
                what = " *"
            print(f"      {pair[0]} -> {pair[1]}: {count:{countWidth}d}{what}")

        print("    Transitions:", totalTransitions)
        print("    Transversions:", totalTransversions)
        print("    TOTAL:", totalTransitions + totalTransversions)


def main() -> None:
    args = makeParser().parse_args()

    if not (args.printOffsets or args.printStats):
        sys.exit("You used both --noOffsets and --noStats, so there is no output!")

    if args.maxIdenticalReadIdsPerSite == 0:
        sys.exit("maxIdenticalReadIdsPerSite cannot be zero.")

    referenceLens = samReferenceLengths(args.samfile)
    referenceId = getReferenceId(args.samfile, set(referenceLens), args.referenceId)
    referenceLen = referenceLens[referenceId]
    referenceSeq = getReferenceSeq(args.fasta, referenceId) if args.fasta else None

    if args.diffsFrom == "reference" and referenceSeq is None:
        sys.exit(
            "If you use --diffsFrom 'reference' (the default), you must use --fasta to "
            "give a FASTA file containing the reference sequence. Else use --diffsFrom "
            "'commonest' to calculate mutations based on the most-common nucleotide "
            "at each genome site."
        )

    printTSV = csv.writer(sys.stdout, delimiter=args.sep).writerow if args.tsv else None
    printRow = printTSV or print

    if printTSV:
        # Print the TSV header.
        printTSV(
            ("Site", "Reference", "Depth")
            + tuple(BASES)
            + ("Base", "nBases", "Mutations", "Variability", "Entropy")
            + tuple(f"{pair[0]}->{pair[1]}" for pair in TRANSITIONS + TRANSVERSIONS)
            + ("Transitions", "Transversions"),
        )

    depths = []
    readIds = set()
    mutationCount = 0
    # I don't use a collections.Counter in the following because they are very slow.
    baseCounts = defaultdict(int)
    mutationPairCounts = defaultdict(int)
    start = time()
    startGMT = gmt()
    totalRunTime = maxRunTime = 0.0
    startSites = list(range(0, referenceLen, args.window))

    with ProcessPoolExecutor(max_workers=args.workers) as executor:
        for result in executor.map(
            processWindowColumns,
            repeat(args.samfile),
            startSites,
            repeat(args.window),
            repeat(args.minDepth),
            repeat(args.maxDepth),
            repeat(args.minBaseQuality),
            repeat(args.minMapQuality),
            repeat(args.minSiteMutationRate),
            repeat(args.maxSiteMutationRate),
            repeat(args.minSiteMutationCount),
            repeat(args.minEntropy),
            repeat(args.maxEntropy),
            repeat(args.minBases),
            repeat(args.maxIdenticalReadIdsPerSite),
            repeat(args.printOffsets),
            repeat(args.verbosity),
            repeat(referenceId),
            repeat(referenceSeq),
            repeat(bool(printTSV)),
            repeat(start),
            repeat(args.diffsFrom),
        ):
            waited = result["startTime"] - start
            runTime = result["stopTime"] - result["startTime"]
            totalRunTime += runTime
            if runTime > maxRunTime:
                maxRunTime = runTime

            if args.verbosity > 1:
                from_ = result["startSite"]
                to = min(from_ + args.window, referenceLen)
                print(
                    f"Genome region {from_ + 1}-{to} processed in {runTime:.2f} "
                    f"seconds (waited {waited:.2f} seconds to start).",
                    file=sys.stderr,
                )

            if args.printOffsets:
                for row in result["rows"]:
                    printRow(row)

            readIds.update(result["readIds"])
            depths.extend(result["depths"])
            for base, count in result["baseCounts"].items():
                baseCounts[base] += count
            for pair, count in result["mutationPairCounts"].items():
                mutationPairCounts[pair] += count
            mutationCount += result["mutationCount"]

    stop = time()
    elapsed = int(stop - start)

    if args.verbosity > 0:
        try:
            speedup = totalRunTime / (stop - start)
        except ZeroDivisionError:
            speedup = 0.0

        print(
            f"Finished in {elapsed} seconds. Total serial (i.e., non parallel) "
            f"processing would have been {int(totalRunTime)} seconds. The slowest "
            f"process took {maxRunTime:.2f} seconds. Speed up = {speedup:.2f}",
            file=sys.stderr,
        )

    if args.printStats:
        printStats(
            args.statsFile,
            args.samfile,
            args.minDepth,
            args.maxDepth,
            args.diffsFrom,
            args.minBases,
            args.maxIdenticalReadIdsPerSite,
            args.minBaseQuality,
            args.minMapQuality,
            referenceSeq,
            referenceId,
            referenceLen,
            startGMT,
            elapsed,
            readIds,
            depths,
            args.minSiteMutationRate,
            args.maxSiteMutationRate,
            args.minSiteMutationCount,
            args.minEntropy,
            args.maxEntropy,
            baseCounts,
            mutationPairCounts,
        )


if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""
Read simplified JSON BLAST records and report the elapsed time.
"""

import sys
from time import time

from dark.blast.conversion import JSONRecordsReader

if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("Usage: %s file.json" % sys.argv[0], file=sys.stderr)
        sys.exit(1)
    else:
        start = time()
        jsonReader = JSONRecordsReader(sys.argv[1])
        records = jsonReader.records()
        for count, record in enumerate(records, start=1):
            pass
        elapsed = time() - start
        print(
            "Read %d JSON BLAST records in %.3f secs (%.0f records/sec)"
            % (count, elapsed, float(count) / float(elapsed))
        )



#!/usr/bin/env python

"""
Given a BLAST or DIAMOND JSON output files, the corresponding FASTA (or FASTQ)
sequence files, and filtering criteria, produce a summary of matched titles
and (optionally) an alignment panel.

Run with --help for help.
"""

import argparse
import os
import sys
from collections import defaultdict
from itertools import chain
from json import dump
from typing import Union

# It's not clear that the PDF backend is the right choice here, but it
# works (i.e., the generation of PNG images works fine).
import matplotlib

matplotlib.use("PDF")

# These imports are here because dark.graphics imports matplotlib.pyplot
# and we need to set the matplotlib backend (see above) before that import
# happens. So please don't move these imports higher in this file.
from dark.blast.alignments import BlastReadsAlignments
from dark.civ.graphics import DEFAULT_LOG_LINEAR_X_AXIS_BASE, alignmentPanelHTML
from dark.civ.proteins import SqliteIndex
from dark.diamond.alignments import DiamondReadsAlignments
from dark.fasta import FastaReads
from dark.fastq import FastqReads
from dark.titles import TitlesAlignments
from dark.utils import numericallySortFilenames


def parseColors(colors, args):
    """
    Parse read id color specification.

    @param colors: A C{list} of C{str}s. Each item is of the form, e.g.,
        'green X Y Z...', where each of X, Y, Z, ... etc. is either a read
        id or the name of a FASTA or FASTQ file containing reads whose ids
        should be displayed with the corresponding color. Note that if read
        ids contain spaces you will need to use the latter (i.e. FASTA/Q file
        name) approach because C{args.colors} is split on whitespace.
    @param args: The argparse C{Namespace} instance holding the other parsed
        command line arguments.
    @return: A C{dict} whose keys are colors and whose values are sets of
        read ids.
    """
    result = defaultdict(set)
    for colorInfo in colors:
        readIds = colorInfo.split()
        color = readIds.pop(0)
        for readId in readIds:
            if os.path.isfile(readId):
                filename = readId
                if args.fasta:
                    reads = FastaReads(filename)
                else:
                    reads = FastqReads(filename)
                for read in reads:
                    result[color].add(read.id)
            else:
                result[color].add(readId)
    return result


if __name__ == "__main__":
    # We do not use the addFASTACommandLineOptions and
    # parseFASTACommandLineOptions utility functions below because we allow
    # multiple FASTA or FASTQ files on the command line, which we specify
    # by --fasta and --fastq. And those names clash with the option names
    # used by those utility functions.

    parser = argparse.ArgumentParser(
        description="Produce an alignment panel in HTML",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        epilog=(
            "Given BLAST or DIAMOND JSON output files, the "
            "corresponding FASTA (or FASTQ) sequence files, and "
            "filtering criteria, produce an alignment panel."
        ),
    )

    parser.add_argument(
        "--proteinGenomeDatabase",
        required=True,
        help=(
            "The filename of an Sqlite3 database holding protein and "
            "genome information, as built by make-protein-database.py"
        ),
    )

    parser.add_argument(
        "--earlyExit",
        action="store_true",
        help=(
            "If True, just print the number of interesting matches, but do "
            "not create the alignment panel."
        ),
    )

    parser.add_argument(
        "--matcher",
        default="blast",
        choices=("blast", "diamond"),
        help="The matching algorithm that was used to produce the JSON.",
    )

    parser.add_argument(
        "--json",
        metavar="JSON-file",
        nargs="+",
        action="append",
        required=True,
        help="the JSON file(s) of BLAST or DIAMOND output.",
    )

    # A mutually exclusive group for either FASTA or FASTQ files.
    group = parser.add_mutually_exclusive_group(required=True)

    group.add_argument(
        "--fasta",
        metavar="FASTA-file",
        nargs="+",
        action="append",
        help=(
            "the FASTA file(s) of read sequences that were given to BLAST or DIAMOND."
        ),
    )

    group.add_argument(
        "--fastq",
        metavar="FASTQ-file",
        nargs="+",
        action="append",
        help=(
            "the FASTQ file(s) of read sequences that were given to BLAST or DIAMOND."
        ),
    )

    # Args for filtering on ReadsAlignments.
    parser.add_argument(
        "--minStart",
        type=int,
        help="Reads that start before this subject offset should not be shown.",
    )

    parser.add_argument(
        "--maxStop",
        type=int,
        help="Reads that end after this subject offset should not be shown.",
    )

    parser.add_argument(
        "--oneAlignmentPerRead",
        action="store_true",
        help="If True, only keep the best alignment for each read.",
    )

    parser.add_argument(
        "--maxAlignmentsPerRead",
        type=int,
        help=(
            "Reads with more than this many alignments will be elided. Pass "
            "zero to only keep reads with no matches (alignments)."
        ),
    )

    parser.add_argument(
        "--scoreCutoff",
        type=float,
        help=("A float score. Matches with scores worse than this will be ignored."),
    )

    parser.add_argument(
        "--percentageIdenticalCutoff",
        type=float,
        help=(
            "A float percentage identity (0.0 to 100.0). Matches with "
            "percent identity scores less than this will be ignored."
        ),
    )

    parser.add_argument(
        "--percentagePositiveCutoff",
        type=float,
        help=(
            "A float percentage (0.0 to 100.0). Matches with percent "
            "positive scores less than this will be ignored. A single AA "
            "match is considered positive (by DIAMOND) if its entry in the "
            "substitution (scoring) matrix is positive. DIAMOND uses "
            "the BLOSUM62 matrix by default."
        ),
    )

    parser.add_argument(
        "--maxHspsPerHit",
        type=int,
        help="A numeric max number of HSPs to show for each hit on hitId.",
    )

    parser.add_argument(
        "--whitelist",
        nargs="+",
        action="append",
        help="sequence titles that should be whitelisted",
    )

    parser.add_argument(
        "--whitelistFile",
        help=("A file containing sequence titles (one per line) to be whitelisted"),
    )

    parser.add_argument(
        "--blacklist",
        nargs="+",
        action="append",
        help="sequence titles that should be blacklisted",
    )

    parser.add_argument(
        "--blacklistFile",
        help=("A file containing sequence titles (one per line) to be blacklisted"),
    )

    parser.add_argument("--titleRegex", help="a regex that sequence titles must match.")

    parser.add_argument(
        "--negativeTitleRegex", help="a regex that sequence titles must not match."
    )

    parser.add_argument(
        "--truncateTitlesAfter",
        help=(
            "a string that titles will be truncated beyond. If the "
            "truncated version of a title has already been seen, "
            "that title will be skipped."
        ),
    )

    parser.add_argument(
        "--minSequenceLen", type=int, help="sequences of lesser length will be elided."
    )

    parser.add_argument(
        "--maxSequenceLen", type=int, help="sequences of greater length will be elided."
    )

    parser.add_argument(
        "--taxonomy",
        help=(
            "a string of the taxonomic group on which should be "
            'filtered. eg "Vira" will filter on viruses.'
        ),
    )

    # Args for filtering on TitlesAlignments.
    parser.add_argument(
        "--minMatchingReads",
        type=int,
        help="sequences that are matched by fewer reads will be elided.",
    )

    parser.add_argument(
        "--maxMatchingReads",
        type=int,
        help="sequences that are matched by more reads will be elided.",
    )

    parser.add_argument(
        "--minMedianScore",
        type=float,
        help=(
            "sequences that are matched with a median score that is "
            "worse will be elided."
        ),
    )

    parser.add_argument(
        "--withScoreBetterThan",
        type=float,
        help=(
            "sequences that are matched without at least one score "
            "at least this good will be elided."
        ),
    )

    parser.add_argument(
        "--minNewReads",
        type=float,
        help=(
            "The fraction of its reads by which a new read set must differ "
            "from all previously seen read sets in order to be considered "
            "acceptably different."
        ),
    )

    parser.add_argument(
        "--maxTitles",
        type=int,
        help=(
            "The maximum number of titles to keep. If more titles than "
            "this result from the filtering, titles will be sorted "
            "(according to the --sortOn value) and only the best will be "
            "retained."
        ),
    )

    parser.add_argument(
        "--minCoverage",
        type=float,
        help=(
            "The (0.0 to 1.0) minimum fraction of a subject sequence that "
            "must be matched by at least one read."
        ),
    )

    # Args for the alignment panel
    parser.add_argument(
        "--sortOn",
        default="maxScore",
        choices=("maxScore", "medianScore", "readCount", "length", "title"),
        help="The attribute to sort subplots on.",
    )

    parser.add_argument(
        "--rankValues",
        action="store_true",
        help=(
            "If True, display reads with a Y axis coord that is the rank of the score."
        ),
    )

    parser.add_argument(
        "--outputDir",
        required=True,
        help="Specifies a directory to write the HTML summary to. Required.",
    )

    parser.add_argument(
        "--color",
        action="append",
        help=(
            "a string which has a color as the first element and read ids "
            "and/or FASTA file names as the following element(s), separated "
            "by spaces."
        ),
    )

    parser.add_argument(
        "--equalizeXAxes",
        action="store_true",
        help=(
            "If True, all alignment graphs will have their X axes drawn "
            "with the same range."
        ),
    )

    parser.add_argument(
        "--xRange",
        default="subject",
        choices=("reads", "subject"),
        help=(
            "Set the X axis range to show either the subject or the extent "
            "of the reads that hit the subject."
        ),
    )

    parser.add_argument(
        "--logLinearXAxis",
        action="store_true",
        help=(
            "If True, convert read offsets so that empty regions in the "
            "alignment panel plots will only be as wide as their logged "
            "actual values"
        ),
    )

    parser.add_argument(
        "--logBase",
        type=float,
        default=DEFAULT_LOG_LINEAR_X_AXIS_BASE,
        help="The base of the logarithm to use if logLinearXAxis is True",
    )

    parser.add_argument(
        "--showFeatures",
        action="store_true",
        help=(
            "If specified, look up features for the individual images in "
            "the alignment panel."
        ),
    )

    parser.add_argument(
        "--sortFilenames",
        action="store_true",
        help=(
            "If specified, the JSON and FASTA/Q file names will be "
            "processed in sorted order. The sorting is based on finding "
            "a numerical prefix in the filename. This can be useful when "
            "processing output files produced by systems like HTCondor, "
            "which makes files with names like 1.out, 10.out, etc. that do "
            "not sort properly and so cannot conveniently be given to this "
            "program along with a single FASTA/Q file (because the order of "
            "the results in the files from HTCondor does not match the "
            "order of sequences in the FASTA/Q file."
        ),
    )

    parser.add_argument(
        "--titlesJSONFile",
        help=(
            "Give a file name for JSON holding information about titles to "
            "be written to"
        ),
    )

    parser.add_argument(
        "--maxBluePlotReads",
        type=int,
        help=(
            "The maximum number of reads to display in a blue plot. Note that "
            "because addQueryLines is unconditionally True (in the calls to "
            "alignmentGraph in alignmentPanelHTML in src/dark/graphics.py), "
            "the number of lines drawn in the plots will be twice the value "
            "specified by --maxBluePlotReads (one line for the gray 'whiskers' "
            "drawn first for the non-matching region, then one line for the "
            "famous blue-plot blue line that shows the actually matched part "
            "of each read)."
        ),
    )

    args = parser.parse_args()

    # Flatten lists of lists that we get from using both nargs='+' and
    # action='append'. We use both because it allows people to use (e.g.)
    # --json on the command line either via "--json file1 --json file2" or
    # "--json file1 file2", or a combination of these. That way it's not
    # necessary to remember which way you're supposed to use it and you also
    # can't be hit by the subtle problem encountered in
    # https://github.com/acorg/dark-matter/issues/453
    jsonFiles = list(chain.from_iterable(args.json))
    whitelist = set(chain.from_iterable(args.whitelist)) if args.whitelist else None
    blacklist = set(chain.from_iterable(args.blacklist)) if args.blacklist else None

    # TODO: Add a --readClass command-line option in case we want to
    # process FASTA containing AA sequences.
    reads: Union[FastaReads, FastqReads]

    if args.fasta:
        if args.sortFilenames:
            files = numericallySortFilenames(
                list(map(str, chain.from_iterable(args.fasta)))
            )
        else:
            files = list(map(str, (chain.from_iterable(args.fasta))))
        reads = FastaReads(files)
    else:
        if args.sortFilenames:
            files = numericallySortFilenames(
                list(map(str, chain.from_iterable(args.fastq)))
            )
        else:
            files = list(chain.from_iterable(args.fastq))
        reads = FastqReads(files)

    readsAlignments: Union[BlastReadsAlignments, DiamondReadsAlignments]

    if args.matcher == "blast":
        readsAlignments = BlastReadsAlignments(
            reads,
            jsonFiles,
            sqliteDatabaseFilename=args.sqliteDatabaseFilename,
            sortBlastFilenames=args.sortFilenames,
        )
    else:
        # Must be 'diamond' (due to parser.add_argument 'choices' argument).
        assert args.matcher == "diamond"

        readsAlignments = DiamondReadsAlignments(
            reads, jsonFiles, sortFilenames=args.sortFilenames
        )

    readsAlignments.filter(
        maxAlignmentsPerRead=args.maxAlignmentsPerRead,
        minSequenceLen=args.minSequenceLen,
        maxSequenceLen=args.maxSequenceLen,
        minStart=args.minStart,
        maxStop=args.maxStop,
        oneAlignmentPerRead=args.oneAlignmentPerRead,
        maxHspsPerHit=args.maxHspsPerHit,
        scoreCutoff=args.scoreCutoff,
        percentageIdenticalCutoff=args.percentageIdenticalCutoff,
        percentagePositiveCutoff=args.percentagePositiveCutoff,
        whitelist=whitelist,
        blacklist=blacklist,
        whitelistFile=args.whitelistFile,
        blacklistFile=args.blacklistFile,
        titleRegex=args.titleRegex,
        negativeTitleRegex=args.negativeTitleRegex,
        truncateTitlesAfter=args.truncateTitlesAfter,
        taxonomy=args.taxonomy,
    )

    titlesAlignments = TitlesAlignments(readsAlignments).filter(
        minMatchingReads=args.minMatchingReads,
        maxMatchingReads=args.maxMatchingReads,
        minMedianScore=args.minMedianScore,
        withScoreBetterThan=args.withScoreBetterThan,
        minNewReads=args.minNewReads,
        maxTitles=args.maxTitles,
        sortOn=args.sortOn,
        minCoverage=args.minCoverage,
    )

    nTitles = len(titlesAlignments)
    print(
        "Found %d interesting title%s." % (nTitles, "" if nTitles == 1 else "s"),
        file=sys.stderr,
    )

    # If we've been asked to save the matched title information in JSON,
    # write it out.
    if args.titlesJSONFile:
        with open(args.titlesJSONFile, "w") as fp:
            dump(
                titlesAlignments.toDict(),
                fp,
                sort_keys=True,
                indent=2,
                separators=(",", ": "),
            )

    if nTitles:
        # Use flush=True on the print of the titles so the output is
        # definitely written out. This is because in some cases the
        # following aligment panel generation takes a long time and exceeds
        # its CPU or memory limit on an HPCS system.  Also, it can be good
        # to see the matching title details while waiting for the panel
        # plots to be generated.
        print(titlesAlignments.tabSeparatedSummary(sortOn=args.sortOn), flush=True)

    if args.earlyExit:
        sys.exit(0)

    if nTitles == 0:
        print(
            "No alignment panel generated due to no matching titles.", file=sys.stderr
        )
        sys.exit(0)

    idList = parseColors(args.color, args) if args.color else None

    alignmentPanelHTML(
        titlesAlignments,
        SqliteIndex(args.proteinGenomeDatabase),
        sortOn=args.sortOn,
        outputDir=args.outputDir,
        idList=idList,
        equalizeXAxes=args.equalizeXAxes,
        xRange=args.xRange,
        logLinearXAxis=args.logLinearXAxis,
        logBase=args.logBase,
        showFeatures=args.showFeatures,
        maxBluePlotReads=args.maxBluePlotReads,
    )



from unittest import TestCase

from dark.reads import DNARead
from dark.windowedIdentity import WindowedIdentity


class TestWindowedIdentity(TestCase):
    """
    Test the WindowedIdentity class.
    """

    def testEmptyFASTA(self):
        """
        If no FASTA sequences are passed, a ValueError must be raised.
        """
        wi = WindowedIdentity([])
        error = "^Empty FASTA input!$"
        self.assertRaisesRegex(ValueError, error, wi.getIdentity, "xxx", 10, 3, 2)

    def testWindowSizeTooSmall(self):
        """
        If the window size is too small, a ValueError must be raised.
        """
        wi = WindowedIdentity(
            [
                DNARead("id", "ACGT"),
            ]
        )
        error = r"^Window size must be positive\. You passed 0\.$"
        self.assertRaisesRegex(ValueError, error, wi.getIdentity, "xxx", 0, 3, 2)

    def testMinWindowSizeNegative(self):
        """
        If the minimum window size is less zero, a ValueError must be raised.
        """
        wi = WindowedIdentity(
            [
                DNARead("id", "ACGT"),
            ]
        )
        error = r"^Minimum window size cannot be negative\. You passed -1\.$"
        self.assertRaisesRegex(ValueError, error, wi.getIdentity, "xxx", 2, -1, 2)

    def testJumpLessThanOne(self):
        """
        If the jump is less than one, a ValueError must be raised.
        """
        wi = WindowedIdentity(
            [
                DNARead("id", "ACGT"),
            ]
        )
        error = r"^Jump must be positive\. You passed 0\.$"
        self.assertRaisesRegex(ValueError, error, wi.getIdentity, "xxx", 2, 1, 0)

    def testStartOffsetTooSmall(self):
        """
        If the start offset is too small, a ValueError must be raised.
        """
        wi = WindowedIdentity(
            [
                DNARead("id", "ACGT"),
            ]
        )
        error = r"^Start offset cannot be negative\. You passed -1\.$"
        self.assertRaisesRegex(
            ValueError, error, wi.getIdentity, "xxx", 2, 3, 2, startOffset=-1
        )

    def testStopOffsetTooSmall(self):
        """
        If the stop offset is too small, a ValueError must be raised.
        """
        wi = WindowedIdentity(
            [
                DNARead("id", "ACGT"),
            ]
        )
        error = r"^Stop offset cannot be negative\. You passed -1\.$"
        self.assertRaisesRegex(
            ValueError, error, wi.getIdentity, "xxx", 2, 3, 2, stopOffset=-1
        )

    def testStopOffsetBeforeStartOffset(self):
        """
        If the stop offset is before the start offset, a ValueError must be raised.
        """
        wi = WindowedIdentity(
            [
                DNARead("id", "ACGT"),
            ]
        )
        error = (
            r"^Stop offset must be greater than start offset. You passed startOffset = 4 "  # noqa: E501
            r"and stopOffset = 3\."
        )
        self.assertRaisesRegex(
            ValueError,
            error,
            wi.getIdentity,
            "xxx",
            2,
            3,
            2,
            startOffset=4,
            stopOffset=3,
        )

    def testStopOffsetEqualsStartOffset(self):
        """
        If the stop offset equals the start offset, a ValueError must be raised.
        """
        wi = WindowedIdentity(
            [
                DNARead("id", "ACGT"),
            ]
        )
        error = (
            r"^Stop offset must be greater than start offset. You passed startOffset = 3 "  # noqa: E501
            r"and stopOffset = 3\."
        )
        self.assertRaisesRegex(
            ValueError,
            error,
            wi.getIdentity,
            "xxx",
            2,
            3,
            2,
            startOffset=3,
            stopOffset=3,
        )

    def testEmptyFirstSequence(self):
        """
        If the first FASTA sequence is empty, a ValueError must be raised.
        """
        wi = WindowedIdentity(
            [
                DNARead("id", ""),
            ]
        )
        error = r"^The first input sequence has length zero!$"
        self.assertRaisesRegex(ValueError, error, wi.getIdentity, "xxx", 10, 3, 2)

    def testNoMatchingReference(self):
        """
        If the reference pattern doesn't match any sequence id, a ValueError must
        be raised.
        """
        wi = WindowedIdentity(
            [
                DNARead("id", "ACGT"),
            ]
        )
        error = (
            r"^No input sequence IDs match the regular expression for the "
            r"reference \('xxx'\)\.$"
        )
        self.assertRaisesRegex(ValueError, error, wi.getIdentity, "xxx", 10, 3, 2)

    def testMultipleMatchingReference(self):
        """
        If the reference pattern matches more than one sequence id, a ValueError must
        be raised.
        """
        wi = WindowedIdentity(
            [
                DNARead("id1", "ACGT"),
                DNARead("id2", "ACGT"),
            ]
        )
        error = (
            r"^2 input sequence ids match the reference pattern! The matching ids are: "
            r"'id1', 'id2'\.$"
        )
        self.assertRaisesRegex(ValueError, error, wi.getIdentity, "id", 10, 3, 2)

    def testNothingToCompareAgainst(self):
        """
        If the include pattern doesn't match any sequence ids, a ValueError must
        be raised.
        """
        wi = WindowedIdentity(
            [
                DNARead("id", "ACGT"),
            ]
        )
        error = r"^No input sequence ids match the include pattern!$"
        self.assertRaisesRegex(
            ValueError, error, wi.getIdentity, "^id$", 10, 3, 2, includeRegex="xxx"
        )

    def testUnequalReadLengths(self):
        """
        If the passed reads are not all the same length, a ValueError must be raised.
        """
        wi = WindowedIdentity(
            [
                DNARead("id1", "A"),
                DNARead("id2", "AA"),
            ]
        )
        error = (
            r"^Sequence number 2, with id 'id2', has length 2 which is not equal to "
            r"the previous input sequence length\(s\) of 1\.$"
        )
        self.assertRaisesRegex(ValueError, error, wi.getIdentity, "^id$", 10, 3, 2)

    def testPerfectMatchOneIncluded(self):
        """
        If the reference sequence matches the single included sequence perfectly,
        we must get back identities that are all 1.0
        """
        ref = DNARead("reference", "ACGT")
        other = DNARead("another", "ACGT")
        wi = WindowedIdentity([ref, other])

        reference, starts, identity = wi.getIdentity("ref", 10, 3, 2)
        self.assertIs(ref, reference)
        self.assertEqual([0], starts)
        self.assertEqual({other: [1.0]}, identity)

    def testPerfectMatchOneIncludedStartAtOne(self):
        """
        If the reference sequence matches the single included sequence perfectly,
        we must get back identities that are all 1.0 and the starts list must begin
        at the passed value (of one).
        """
        ref = DNARead("reference", "ACGT")
        other = DNARead("another", "ACGT")
        wi = WindowedIdentity([ref, other])

        reference, starts, identity = wi.getIdentity("ref", 10, 3, 2, startOffset=1)
        self.assertIs(ref, reference)
        self.assertEqual([1], starts)
        self.assertEqual({other: [1.0]}, identity)

    def testPerfectMatchOneIncludedStartAtOneFirstCharDiffers(self):
        """
        If the reference sequence matches the single included sequence perfectly,
        except for its first character, we must get back identities that are all
        1.0 and the starts list must begin at the passed value (of one).
        """
        ref = DNARead("reference", "CCGT")
        other = DNARead("another", "ACGT")
        wi = WindowedIdentity([ref, other])

        reference, starts, identity = wi.getIdentity("ref", 10, 3, 2, startOffset=1)
        self.assertIs(ref, reference)
        self.assertEqual([1], starts)
        self.assertEqual({other: [1.0]}, identity)

    def testPerfectMatchOneIncludedIgnoreFirstAndLastChars(self):
        """
        If the reference sequence matches the single included sequence perfectly,
        except for its first and last characters, we must get back identities that
        are all 1.0 and the starts list must begin at the passed value (of one).
        """
        ref = DNARead("reference", "CCGC")
        other = DNARead("another", "ACGT")
        wi = WindowedIdentity([ref, other])

        reference, starts, identity = wi.getIdentity(
            "ref", 10, 2, 2, startOffset=1, stopOffset=3
        )
        self.assertIs(ref, reference)
        self.assertEqual([1], starts)
        self.assertEqual({other: [1.0]}, identity)

    def testOneMatchOneMismatchIgnoreFirstAndLastChars(self):
        """
        If the reference sequence matches the single included sequence in the second
        character and not the third and the first and last characters are ignored and
        the window, min window, and jump size are all one, we must get back identities
        of 1.0 and 0.0 and the starts list must have 1 and 2.
        """
        ref = DNARead("reference", "CCCC")
        other = DNARead("another", "ACGT")
        wi = WindowedIdentity([ref, other])

        reference, starts, identity = wi.getIdentity(
            "ref", 1, 1, 1, startOffset=1, stopOffset=3
        )
        self.assertIs(ref, reference)
        self.assertEqual([1, 2], starts)
        self.assertEqual({other: [1.0, 0.0]}, identity)

    def testMismatchOneIncluded(self):
        """
        If the reference sequence mismatches the single included sequence completely,
        we must get back identities that are all 0.0
        """
        ref = DNARead("reference", "TTTT")
        other = DNARead("another", "GGGG")
        wi = WindowedIdentity([ref, other])

        reference, starts, identity = wi.getIdentity("ref", 10, 3, 2)
        self.assertIs(ref, reference)
        self.assertEqual([0], starts)
        self.assertEqual({other: [0.0]}, identity)

    def testHalfMatchOneIncluded(self):
        """
        If the reference sequence matches the single included sequence in half its
        locations we must get back identities that are all 0.5
        """
        ref = DNARead("reference", "AATT")
        other = DNARead("another", "AAGG")
        wi = WindowedIdentity([ref, other])

        reference, starts, identity = wi.getIdentity("ref", 10, 3, 2)
        self.assertIs(ref, reference)
        self.assertEqual([0], starts)
        self.assertEqual({other: [0.5]}, identity)

    def testAmbiguousNonStrictMatch(self):
        """
        If the reference sequence only matches the single included sequence via
        an ambiguous code (just one of two nucelotides matching) we must get back
        an identity of 0.5.
        """
        ref = DNARead("reference", "TT")
        other = DNARead("another", "GW")
        wi = WindowedIdentity([ref, other])

        reference, starts, identity = wi.getIdentity("ref", 10, 1, 2, strict=False)
        self.assertIs(ref, reference)
        self.assertEqual([0], starts)
        self.assertEqual({other: [0.5]}, identity)

    def testOneIncludedWindowSizeTwo(self):
        """
        If the reference sequence matches the single included sequence perfectly
        in its first two positions, 50% in the second two, and not at all in the final
        two (with a window and jump both equal to two), we must get back identities of
        1.0, 0.5, and 0.0
        """
        ref = DNARead("reference", "AATTCC")
        other = DNARead("another", "AATGAA")
        wi = WindowedIdentity([ref, other])

        reference, starts, identity = wi.getIdentity("ref", 2, 1, 2)
        self.assertIs(ref, reference)
        self.assertEqual([0, 2, 4], starts)
        self.assertEqual({other: [1.0, 0.5, 0.0]}, identity)

    def testOneIncludedWindowSizeTwoJumpOne(self):
        """
        If the reference sequence matches the single included sequence perfectly
        in its first two positions, 50% in the second two, and not at all in the final
        two (with a window size two, a min window of size two, and jump of one), we must
        get back the expected identities.
        """
        ref = DNARead("reference", "AATTCC")
        other = DNARead("another", "AATGAA")
        wi = WindowedIdentity([ref, other])

        reference, starts, identity = wi.getIdentity("ref", 2, 2, 1)
        self.assertIs(ref, reference)
        self.assertEqual([0, 1, 2, 3, 4], starts)
        self.assertEqual({other: [1.0, 1.0, 0.5, 0.0, 0.0]}, identity)

    def testOneIncludedWindowSizeTwoJumpOneMinwindowOne(self):
        """
        If the reference sequence matches the single included sequence perfectly
        in its first two positions, 50% in the second two, and not at all in the final
        two (with a window size two, a min window of one and a jump of one), we must
        get back the expected identities.
        """
        ref = DNARead("reference", "AATTCC")
        other = DNARead("another", "AATGAA")
        wi = WindowedIdentity([ref, other])

        reference, starts, identity = wi.getIdentity("ref", 2, 1, 1)
        self.assertIs(ref, reference)
        self.assertEqual([0, 1, 2, 3, 4, 5], starts)
        self.assertEqual({other: [1.0, 1.0, 0.5, 0.0, 0.0, 0.0]}, identity)

    def testOneIncludedWindowSizeTwoIgnoreNonMatchingSequence(self):
        """
        If the reference sequence matches the single included sequence perfectly
        in its first two positions, 50% in the second two, and not at all in the final
        two (with a window and jump both equal to two), we must get back identities of
        1.0, 0.5, and 0.0. An additional sequence with a non-matching id must be
        ignored if it does not match the include regex.
        """
        ref = DNARead("reference", "AATTCC")
        other = DNARead("another", "AATGAA")
        ignored = DNARead("dummy", "AATGAA")
        wi = WindowedIdentity([ref, other, ignored])

        reference, starts, identity = wi.getIdentity("ref", 2, 1, 2, includeRegex="her")
        self.assertIs(ref, reference)
        self.assertEqual([0, 2, 4], starts)
        self.assertEqual({other: [1.0, 0.5, 0.0]}, identity)

    def testDarkMatterIssue802(self):
        """
        Trying to reproduce the failure seen in
        https://github.com/acorg/dark-matter/issues/802 with the exact sequence ids
        that caused that issue.
        """
        ids = """
            NC_003977 FJ899792 JN642140 GQ477453 GQ477455 JN642160 JN642163
            JN688710 JN688711 GQ922005 HE974378 KJ470893 KJ470896 KJ470898
            FJ904430 FJ904436 AB033559 AB048701 AB048702 AB188243 AB210818
            AM494716 AY796031 AY902768 DQ315779 X80925 FJ904399 AY721612
            AY741797 AB270543 EU594409 AB109476 AB555496 GQ205377 MG585269
            AKB003 BRE008 BRE026 BRE028 CUN002 I0157 I0161 I0216 I0217 I1344
            KIL044 KRA001 KRA010 MAY017 MLR005 OAI017 ORE002 Petersberg
            RMI002 SED009 SHK001 STR144 ZWE008 AB486012 AY330911 AJ131571
            AY781180 U46935 AJ131567 AF193863 EU155824 AF498266 BNL005 BON020
            CLL005 DER027 GRG036 I0061 I0104 KLE031 Loschbour MIB003 MIB040
            MIB041 MIS002 MKL025 MN2003 MOT001 MPR001 OOH025 PLZ001 SEC006
            SUA002 TRI011 UOO033 UZZ075 UZZ081 WEH008 WEH009 WSN012 I0461
            KVK001 ARM001 ARM0023 ELT005 GGN005 GRG001 HAL15 HOP004 HSL001
            I0100 I0411 I0551 I0795 I0798 I2020 I2031 JAZ001 MUR007 PDA001
            PEN003 UZZ061 VLI016 YUN048 KAP002 MIB011 AB076679 AB116084
            AB453988 AY738142 GQ477499 AY934764 FJ692556 FJ692598 FJ692611
            GQ161813 GQ331046 AB073858 AB033555 AB219429 AP011089 AB073835
            AB287316 AB287318 AB287320 AB287321 DQ463789 DQ463792 AB241117
            DQ993686 AB111946 AB112066 AB112472 DQ089767 X75656 X75665
            AB048704 AF241411 AP011100 AP011102 AP011103 AP011106 AP011108
            FJ899779 AB049609 FJ023669 EU835241 JN315779 SJN001
            Abusir1543 BOO006 BOO008 CHT001 I1321 KBD002 SGR004 AB048705
            CAO009 AY090458 AB116654 FJ657525 AY090455 AY311369 DQ899144
            DQ899146 AB116549 X75663 AF223962 AB166850 AB059660 AB375163
            AY090454 AY090457 JN792922 KUE033 MAP007 PLS004 SJN013 SGR003
            VLI060 AB056513 AB064312 AF405706 HM363593 X75657 X75664 MK5004
            MK5009 PDA003 LEU065 ELT006 HAL05 ISB002 SID005 UZZ099 VTZ011
            I0784 DA45 DA51 NEO597 NEO238 17662 VK19 VK211-D VK223 RISE254
            DA357 DA195 VK408 Sorsum_low_freq VK18 VK161 DA251
            Sorsum_high_freq DA335 KolymaRiver Karsdorf RISE718 DA222 NEO2324
            DA29 VK477 VK211-G NEO721 DA27 DA119 HBV-A1_N4407 HBV-A2_N4879
            HBV-B2_N4214 HBV-C2_N3825 HBV-D1_N4203 HBV-D3_3091 HBV-G
        """.split()

        ref = DNARead("AB562463", "AATTCC")
        others = [DNARead(id_, "AATGAA") for id_ in ids]
        wi = WindowedIdentity([ref] + others)
        regex = "FJ023669|HBV-G|BOO006|SGR004|AB048705|17662|EU835241|RISE718"

        reference, starts, identity = wi.getIdentity(
            "AB562463", 2, 1, 2, includeRegex=regex
        )

        self.assertIs(ref, reference)
        self.assertEqual([0, 2, 4], starts)
        expected = dict.fromkeys(
            [DNARead(id__, "AATGAA") for id__ in set(regex.split("|"))], [1.0, 0.5, 0.0]
        )
        self.assertEqual(expected, identity)



from unittest import TestCase

from dark.codonDistance import codonInformation, findDistance


class TestCodonInformation(TestCase):
    """
    Tests for the codonInformation function.
    """

    def testReturnRightCodon(self):
        """
        The function must return the right codon distances.
        """
        result = codonInformation(["GAT", "GAC"], ["GCC", "GCA"])
        self.assertEqual([("GAC", "GCC")], result[1])
        self.assertEqual([("GAT", "GCC"), ("GAT", "GCA"), ("GAC", "GCA")], result[2])

    def testReturnRightCodonWithTransitionCount(self):
        """
        The function must return the right codon distances and the correct
        number of transitions.
        """
        result = codonInformation(["GAT", "GAC"], ["GCC", "AGT"], countTransitions=True)
        self.assertEqual([("GAC", "GCC", 0)], result[1])
        self.assertEqual([("GAT", "GCC", 1), ("GAT", "AGT", 2)], result[2])
        self.assertEqual([("GAC", "AGT", 3)], result[3])


class TestFindDistance(TestCase):
    """
    Tests for the findDistance function.
    """

    def testFindDistanceThree(self):
        """
        findDistance must return the right distance if codons are different.
        """
        distance = findDistance("ACA", "TGT")
        self.assertEqual(3, distance)

    def testFindDistanceZero(self):
        """
        findDistance must return the right distance if codons are identical.
        """
        distance = findDistance("ACA", "ACA")
        self.assertEqual(0, distance)



#!/usr/bin/env python3

import argparse
import csv
import sys

from dark.reads import (
    Read,
    ReadsInRAM,
    addFASTACommandLineOptions,
    parseFASTACommandLineOptions,
)
from dark.utils import pct


def get_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description=(
            "Find nucleotide sequences in an MSA and extract the matched region."
        ),
    )

    addFASTACommandLineOptions(parser)

    parser.add_argument("--prefix", help="The start of the nucleotide sequence.")

    parser.add_argument("--suffix", help="The end of the nucleotide sequence.")

    parser.add_argument(
        "--id",
        help=(
            "The id of the input sequence in which the prefix and suffix must be "
            "found. If not specified, all input sequences will be examined and "
            "those in which the prefix/suffix are found must all agree."
        ),
    )

    parser.add_argument(
        "--allowUnequalLengths",
        action="store_true",
        help=(
            "Do not exit if all sequences are not the same length. I.e., if this is "
            "not an MSA."
        ),
    )

    parser.add_argument(
        "--allowDistributedMatch",
        action="store_true",
        help=(
            "If --prefix and --suffix are both used, the default behaviour is to "
            "make sure that they are found together in at least one of the input "
            "sequences. If you don't mind if the prefix is found in one sequence "
            "and the suffix in another, use this option to allow it."
        ),
    )

    parser.add_argument(
        "--adjustIds",
        action="store_true",
        help=(
            "Add ' (position X-Y)' to the end of the output sequence ids to indicate "
            "the subsequence range of the result."
        ),
    )

    parser.add_argument(
        "--verbose",
        action="store_true",
        help="Print (to standard error) a summary of total prefixes and suffixes found.",  # noqa: E501
    )

    parser.add_argument(
        "--zeroBased",
        action="store_true",
        help=(
            "If --tsv or --verbose are used, the reported prefix and suffix positions "
            "will refer to zero-based Python style [closed, open) regions."
        ),
    )

    parser.add_argument(
        "--gapCharacter", default="-", help="The 1-letter gap character."
    )

    parser.add_argument(
        "--tsv", help="The (optional) name of a TSV file to write match offsets to."
    )

    return parser.parse_args()


def summarize(
    prefix: str,
    suffix: str,
    reads: ReadsInRAM,
    overallStart: int,
    overallEnd: int,
    adjust: int,
    resultReads: ReadsInRAM,
    details: list[tuple[tuple[int, int], Read]],
    allowUnequalLengths: bool,
    gapCharacter: str,
) -> None:
    nSequences = len(reads)
    inputLengths = set(len(read) for read in reads)

    resultLengths = set(len(read) for read in resultReads)
    assert len(inputLengths) == 1 and len(resultLengths) == 1 or allowUnequalLengths
    ungappedResultLengths = set(
        len(read.sequence.replace(gapCharacter, "")) for read in resultReads
    )

    if len(inputLengths) == 1:
        length = f"{inputLengths.pop()}"
        summary = [f"Read {nSequences} sequences of length {length}."]
    else:
        lengths = ", ".join(map(str, sorted(inputLengths)))
        summary = [f"Read {nSequences} sequences (lengths {lengths}."]

    if prefix:
        summary.append(
            f"The prefix was found at position {overallStart + adjust} ({adjust}-based)."  # noqa: E501
        )
    if suffix:
        summary.append(f"The suffix ends at position {overallEnd} ({adjust}-based).")

    if len(resultLengths) == 1:
        summary.append(f"Extracted alignment region has length {resultLengths.pop()}.")
    else:
        lengths = ", ".join(map(str, sorted(resultLengths)))
        summary.append(f"Extracted aligned region sequences have lengths {lengths}.")

    if len(ungappedResultLengths) == 1:
        summary.append(
            f"Extracted ungapped sequences have length {ungappedResultLengths.pop()}."
        )
    else:
        lengths = ", ".join(map(str, sorted(ungappedResultLengths)))
        summary.append(f"Extracted ungapped sequences have lengths {lengths}.")

    prefixMatches = suffixMatches = 0
    for (start, end), _ in details:
        prefixMatches += start != -1
        suffixMatches += end != -1

    if prefix:
        summary.append(f"Prefix match count: {pct(prefixMatches, nSequences)}.")

    if suffix:
        summary.append(f"Suffix match count: {pct(suffixMatches, nSequences)}.")

    print("\n".join(summary), file=sys.stderr)


def writeTSV(
    resultReads: ReadsInRAM,
    details: list[tuple[tuple[int, int], Read]],
    tsvFile: str,
    adjust: int,
    gapCharacter: str,
) -> None:
    with open(tsvFile, "w") as fp:
        writerow = csv.writer(fp, delimiter="\t").writerow
        writerow(
            (
                "Sequence ID",
                "Prefix offset",
                "Suffix offset",
                "Match count",
                "Ungapped length",
            )
        )

        for read, ((start, end), _) in zip(resultReads, details):
            writerow(
                (
                    read.id,
                    -1 if start == -1 else start + adjust,
                    end,
                    (start != -1) + (end != -1),
                    len(read.sequence.replace(gapCharacter, "")),
                ),
            )


def main():
    args = get_args()

    prefix = args.prefix
    suffix = args.suffix
    adjust = int(not args.zeroBased)
    reads = ReadsInRAM(parseFASTACommandLineOptions(args))

    resultReads, (overallStart, overallEnd), details = reads.extractRegion(
        id_=args.id,
        prefix=prefix,
        suffix=suffix,
        ignoreGaps=True,
        allowUnequalLengths=args.allowUnequalLengths,
    )

    if overallStart != -1 and overallEnd != -1 and overallStart >= overallEnd:
        sys.exit(
            f"The prefix start position ({overallStart + adjust}) is not less "
            f"than the suffix end position ({overallEnd})"
        )

    if prefix and overallStart == -1:
        sys.exit("The prefix was not found in any sequence.")

    if suffix and overallEnd == -1:
        sys.exit("The suffix was not found in any sequence.")

    if prefix and suffix and not args.allowDistributedMatch:
        for (start, end), _ in details:
            if start != -1 and end != -1:
                # The prefix and suffix were both found in this read, so we're good.
                break
        else:
            extra = f"See {args.tsv!r}" if args.tsv else "Re-run with --tsv"
            sys.exit(
                "Simultaneous matches of the prefix and the suffix did not occur in "
                f"any one of the input sequences. {extra} for match details."
            )

    resultReads = ReadsInRAM(resultReads)

    if args.adjustIds:
        for read, ((start, end), _) in zip(resultReads, details):
            start = 0 if start == -1 else start
            end = len(read) if end == -1 else end
            read.id += f" (positions {start + adjust}-{end})"

    resultReads.save(sys.stdout)

    if args.verbose:
        summarize(
            prefix,
            suffix,
            reads,
            overallStart,
            overallEnd,
            adjust,
            resultReads,
            details,
            args.allowUnequalLengths,
            args.gapCharacter,
        )

    if args.tsv:
        writeTSV(resultReads, details, args.tsv, adjust, args.gapCharacter)


if __name__ == "__main__":
    main()



import subprocess
from io import StringIO

from Bio import SeqIO


def getSequence(title, db="nt"):
    """
    @param title: A C{str} sequence title from a BLAST hit. Of the form
        'gi|63148399|gb|DQ011818.1| Description...'.
    @param db: the C{str} name of the BLAST database to search.
    @return: A C{SeqIO.read} instance.
    """
    titleId = title.split(" ", 1)[0]
    fasta = subprocess.check_output(
        ["blastdbcmd", "-entry", titleId, "-db", db]
    ).decode("ascii")
    return SeqIO.read(StringIO(fasta), "fasta")



#!/usr/bin/env python

import argparse
import multiprocessing
import os
import sys
from os.path import exists, join
from tempfile import mkdtemp

import pysam

from dark.bowtie2 import Bowtie2
from dark.process import Executor

DEFAULT_SAMTOOLS_VIEW_FLAGS = (
    pysam.FUNMAP | pysam.FSECONDARY | pysam.FDUP | pysam.FSUPPLEMENTARY
)


def saveStdin(args, e):
    if os.isatty(0):
        print("Reading sequences to match against from stdin.", file=sys.stderr)
    dirname = mkdtemp(prefix="run-bt2-stdin-", dir=args.tempdir)
    if args.tmpChmod:
        e.execute(f"chmod {args.tmpChmod} {dirname}")
    filename = join(dirname, "stdin.fastq")
    count = 0

    with open(filename, "w") as fp:
        for line in sys.stdin:
            fp.write(line)
            count += 1

    if count == 0:
        print("Standard input was empty! Exiting.", file=sys.stderr)
        sys.exit(1)

    return dirname, filename


def processMatch(args, e):
    """
    Run Bowtie2 to find matches.
    """
    if not args.index and not args.reference:
        print("One of --index or --reference must be given.", file=sys.stderr)
        sys.exit(0)

    if args.markDuplicatesPicard or args.callHaplotypesGATK:
        if args.picardJar:
            picardJar = args.picardJar
        else:
            try:
                picardJar = os.environ["PICARD_JAR"]
            except KeyError:
                print(
                    "If you use --markDuplicatesPicard or "
                    "--callHaplotypesGATK, you must give a Picard JAR file "
                    "with --picardJar or else set PICARD_JAR in your "
                    "environment.",
                    file=sys.stderr,
                )
                sys.exit(0)

    bt2 = Bowtie2(
        executor=e,
        threads=(multiprocessing.cpu_count() if args.threads is None else args.threads),
        verboseFp=(sys.stderr if args.verbose else None),
        dryRun=args.dryRun,
        tempdir=args.tempdir,
        tmpChmod=args.tmpChmod,
    )

    bt2.buildIndex(args.index or args.reference)

    if not args.align:
        if args.verbose:
            print("Bowtie2 alignment not done due to --noAlign.", file=sys.stderr)
        print("Bowtie2 temporary directory %r." % bt2.tempdir, file=sys.stderr)
        sys.exit(0)

    if args.fastq1:
        stdinDir, fastq1, fastq2 = None, args.fastq1, args.fastq2
    else:
        stdinDir, fastq1 = saveStdin(args, e)
        fastq2 = None

    if args.out:
        if exists(args.out) and not (args.force or args.dryRun):
            print(
                "Will not overwrite pre-existing output file %r. "
                "Use --force to make me." % args.out,
                file=sys.stderr,
            )
            sys.exit(1)

        if args.indexBAM:
            bai = args.out + ".bai"
            if exists(bai) and not (args.force or args.dryRun):
                print(
                    "Will not overwrite pre-existing output file %r. "
                    "Use --force to make me." % bai,
                    file=sys.stderr,
                )
                sys.exit(1)
            needBAI = True
        else:
            needBAI = False

    if args.callHaplotypesGATK:
        if args.vcfFile:
            for filename in (args.vcfFile, args.vcfFile + ".tbi"):
                if exists(filename) and not args.force:
                    print(
                        "Will not overwrite pre-existing VCF file %r. "
                        "Use --force to make me." % filename,
                        file=sys.stderr,
                    )
                    sys.exit(1)

    if args.callHaplotypesBcftools:
        if args.vcfFile:
            if exists(args.vcfFile) and not args.force:
                print(
                    "Will not overwrite pre-existing VCF file %r. "
                    "Use --force to make me." % args.vcfFile,
                    file=sys.stderr,
                )
                sys.exit(1)

    bt2.align(bowtie2Args=args.bowtie2Args, fastq1=fastq1, fastq2=fastq2)

    indexed = False

    # Sort, remove duplicates, etc. only if there is something in the BAM file.  Apart
    # from a slight efficiency gain, we check this because gatk dies with a typically
    # horrible Java stack dump if the are no reads in a BAM/SAM file.
    if not bt2.isEmpty():
        if args.sort and not (
            args.markDuplicatesPicard
            or args.markDuplicatesGATK
            or args.removePrimersFromBedFile
        ):
            bt2.sort()

        if args.removePrimersFromBedFile:
            bt2.sort()
            bt2.removePrimers(args.removePrimersFromBedFile)

        if args.markDuplicatesPicard:
            bt2.sort()
            bt2.picard(picardJar)

        if args.markDuplicatesGATK:
            bt2.sort(byName=True)
            bt2.markDuplicatesGATK()

        if args.removeDuplicates:
            bt2.removeDuplicates()

        if args.callHaplotypesGATK:
            bt2.makeBAM()
            bt2.indexBAM()
            indexed = True
            bt2.callHaplotypesGATK(
                picardJar=picardJar, vcfFile=args.vcfFile, referenceFasta=args.reference
            )

        if args.callHaplotypesBcftools:
            if not indexed:
                bt2.makeBAM()
                bt2.indexBAM()
                indexed = True
            bt2.callHaplotypesBcftools(
                vcfFile=args.vcfFile, referenceFasta=args.reference
            )

    if args.bam and args.indexBAM and not indexed:
        bt2.makeBAM()
        bt2.indexBAM()
        indexed = True

    if args.out:
        e.execute("mv '%s' '%s'" % (bt2.outputFile(), args.out))
        if needBAI:
            e.execute("mv '%s.bai' '%s.bai'" % (bt2.outputFile(), args.out))
    else:
        if not args.dryRun:
            with open(bt2.outputFile(), "rb") as fp:
                read, write = fp.read, sys.stdout.write
                while True:
                    data = read(2048)
                    if data:
                        write(data)
                    else:
                        break

    if stdinDir:
        # Remove the directory where we stashed standard input.
        e.execute("rm -r '%s'" % stdinDir)

    if args.clean:
        bt2.close()
    else:
        print("Bowtie2 temporary directory %r." % bt2.tempdir, file=sys.stderr)


def processOneIgnore(args, index, count, tempdir, e):
    """
    Process one ignored index.
    """
    if args.verbose:
        print("Preparing to ignore reads matching index %r." % index, file=sys.stderr)

    bt2 = Bowtie2(
        executor=e,
        threads=(multiprocessing.cpu_count() if args.threads is None else args.threads),
        verboseFp=(sys.stderr if args.verbose else None),
        dryRun=args.dryRun,
        tempdir=tempdir,
        tmpChmod=args.tmpChmod,
    )

    bt2.buildIndex(index)

    unAlignedFile = join(tempdir, "unaligned-%d" % count)

    if args.fastq1 and args.fastq2:
        bt2.align(
            bowtie2Args="%s --un-conc-gz %s" % (args.bowtie2Args, unAlignedFile),
            fastq1=args.fastq1,
            fastq2=args.fastq2,
            discardSAM=True,
        )

        for i in "1", "2":
            src = unAlignedFile + "." + i
            dst = src + ".gz"
            e.execute("mv '%s' '%s'" % (src, dst))
            setattr(args, "fastq" + i, dst)

        if not e.dryRun:
            assert exists(args.fastq1)
            assert exists(args.fastq2)
    else:
        bt2.align(
            bowtie2Args="%s --un-gz %s" % (args.bowtie2Args, unAlignedFile),
            fastq1=args.fastq1,
            discardSAM=True,
        )

        e.execute("mv '%s' '%s.gz'" % (unAlignedFile, unAlignedFile))
        args.fastq1 = unAlignedFile + ".gz"
        if not e.dryRun:
            assert exists(args.fastq1)

    if args.clean:
        bt2.close()
    else:
        print(
            "Bowtie2 temporary ignore index directory %r." % bt2.tempdir,
            file=sys.stderr,
        )


def processIgnores(args, e):
    """
    Ignore the indices in args.ignoredIndices
    """
    if e.dryRun:
        tempdir = "/tmp/ignores"
    else:
        tempdir = mkdtemp(prefix="bt2-ignores-", dir=args.tempdir)
        if args.tmpChmod:
            e.execute(f"chmod {args.tmpChmod} {tempdir}")

    for count, ignoreIndex in enumerate(args.ignoredIndices):
        processOneIgnore(args, ignoreIndex, count, tempdir, e)
    return tempdir


def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            "Run bowtie2 on a FASTA file. Optionally convert the "
            "result to BAM, sorting, and indexing."
        ),
    )

    parser.add_argument(
        "--index",
        help=(
            "Either: an accession number, a filename or the name of a "
            "pre-existing bowtie2 index (created with bowtie2-build). If "
            "not given and --reference is used, the reference will be "
            "used to build a bowtie2 index."
        ),
    )

    parser.add_argument(
        "--ignoreIndex",
        action="append",
        dest="ignoredIndices",
        help=(
            "Either: an accession number, a filename or the name of a "
            "pre-existing bowtie2 index (created with bowtie2-build). "
            "Reads matching this index will be ignored. May be repeated."
        ),
    )

    parser.add_argument(
        "--fastq1",
        "-1",
        help=(
            "The FASTQ reads to match against the bowtie2 index given by "
            "--index. Also use --fast2 if you have paired reads. "
            "If not given, single-end FASTQ reads will be read from "
            "standard input."
        ),
    )

    parser.add_argument(
        "--fastq2",
        "-2",
        help=(
            "The FASTQ reads to match against the bowtie2 index given by "
            "--index. Use this with --fastq1 to specify the mate "
            "file for paired-end reads."
        ),
    )

    parser.add_argument(
        "--bowtie2Args",
        default="--no-unal",
        help=(
            "Extra arguments to be passed to Bowtie2 (use --threads to "
            "specify a thread count)."
        ),
    )

    parser.add_argument(
        "--samtoolsViewArgs",
        default="-F %d -q 30" % DEFAULT_SAMTOOLS_VIEW_FLAGS,
        help="Arguments to be passed to samtools view to create the BAM file.",
    )

    parser.add_argument(
        "--tempdir",
        help=(
            "The temporary directory to use. If not specified, the value "
            "of the TMPDIR environment variable (if any) is used, or else "
            "/tmp."
        ),
    )

    parser.add_argument(
        "--out",
        "-o",
        help=(
            "The output file name. If not given, the resulting SAM or BAM "
            "will be written to standard output will be used."
        ),
    )

    parser.add_argument(
        "--reference",
        help=(
            "The reference FASTA file for use with --callHaplotypesGATK and "
            "--callHaplotypesBcftools. This will be used to build a Bowtie2 "
            "index if --index is not given."
        ),
    )

    parser.add_argument(
        "--vcfFile",
        help=(
            "The file to write VCF info to if --callHaplotypesGATK or "
            "--callHaplotypesBcftools are used."
        ),
    )

    parser.add_argument(
        "--markDuplicatesGATK",
        action="store_true",
        help=(
            "Use GATK to mark duplicates. See "
            "https://gatk.broadinstitute.org for details on GATK."
        ),
    )

    parser.add_argument(
        "--markDuplicatesPicard",
        action="store_true",
        help=(
            "Use Picard to mark duplicates. See "
            "https://github.com/broadinstitute/picard for details on "
            "Picard."
        ),
    )

    parser.add_argument(
        "--picardJar",
        help=(
            "The path to the Picard jar file. See "
            "https://github.com/broadinstitute/picard for details on "
            "Picard."
        ),
    )

    parser.add_argument(
        "--tmpChmod",
        help=(
            "A chmod string for setting the permission on the temporary "
            "directory created. This will be passed to chmod(1), so you "
            'could specify "g+rwx", for example.'
        ),
    )

    parser.add_argument(
        "--removeDuplicates",
        action="store_true",
        help=(
            "Remove duplicates from the resulting SAM/BAM file. Best used "
            "in combination with an option that marks duplicates, such as "
            "--markDuplicatesGATK."
        ),
    )

    parser.add_argument(
        "--verbose",
        action="store_true",
        help=(
            "Print a description of commands as they are (or would be, if "
            "--dryRun is used) executed."
        ),
    )

    parser.add_argument(
        "--log",
        action="store_true",
        help=(
            "Show a log of commands that were (or would be, if --dryRun is "
            "used) executed."
        ),
    )

    parser.add_argument(
        "--threads",
        type=int,
        help="The number of threads to use when running bowtie2 commands.",
    )

    parser.add_argument(
        "--noAlign",
        action="store_false",
        dest="align",
        help="Do not align with Bowtie2, just build an index.",
    )

    parser.add_argument(
        "--noBAM", action="store_false", dest="bam", help="Do not convert SAM to BAM."
    )

    parser.add_argument(
        "--noSort", action="store_false", dest="sort", help="Do not sort the BAM."
    )

    parser.add_argument(
        "--noIndexBAM",
        action="store_false",
        dest="indexBAM",
        help="Do not index the BAM file.",
    )

    parser.add_argument(
        "--noClean",
        action="store_false",
        dest="clean",
        help="Do not remove intermediate files or the temporary directory.",
    )

    parser.add_argument(
        "--force", action="store_true", help="Overwrite pre-existing output file."
    )

    parser.add_argument(
        "--dryRun",
        action="store_true",
        help="Do not run commands, just print what would be done.",
    )

    haplotypeCaller = parser.add_mutually_exclusive_group()

    haplotypeCaller.add_argument(
        "--callHaplotypesGATK",
        action="store_true",
        help=(
            "Use GATK to call haplotypes. See "
            "https://gatk.broadinstitute.org for details on GATK."
        ),
    )

    haplotypeCaller.add_argument(
        "--callHaplotypesBcftools",
        action="store_true",
        help="Use bcftools call to call haplotypes.",
    )

    parser.add_argument(
        "--removePrimersFromBedFile",
        help=(
            "If a bed file with Primers is specified, the Primers "
            "will be soft-clipped from the bam file using iVar"
        ),
    )

    args = parser.parse_args()

    if args.indexBAM and not args.bam:
        print(
            "The --indexBAM option only makes sense if you do not use --noBAM.",
            file=sys.stderr,
        )
        sys.exit(1)

    e = Executor(args.dryRun)

    if args.tempdir is None:
        args.tempdir = os.environ.get("TMPDIR", "/tmp")

    if args.ignoredIndices:
        ignoresDir = processIgnores(args, e)

    processMatch(args, e)

    if args.ignoredIndices:
        if args.clean:
            e.execute(f"rm -r {ignoresDir!r}")
        else:
            print(
                "Temporary directory with non-ignored inputs {ignoresDir!r}.",
                file=sys.stderr,
            )

    if args.dryRun or args.log:
        print("\n".join(e.log), file=sys.stderr)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import argparse
import sys

from dark.sam import ReadsSummary


def main(args):
    """
    Print SAM/BAM file reference read counts.

    @param args: An argparse namespace with information about parsed
        command-line options.
    """
    if args.topReferenceIdsFile and args.sortBy not in ("count", "coverage"):
        print(
            "--topReferenceIdsFile only makes sense when sorting by count or coverage",
            file=sys.stderr,
        )
        sys.exit(1)

    summary = ReadsSummary(args.samFile)

    print(
        summary.toString(
            excludeZeroes=args.excludeZeroes,
            excludeIfNoAdditional=args.excludeIfNoAdditional,
            sortBy=args.sortBy,
        )
    )

    # Write out the sorted read ids of the top reference.
    if args.topReferenceIdsFile:
        with open(args.topReferenceIdsFile, "w") as fp:
            print("\n".join(sorted(summary.sortedReferences[0].readIds)), file=fp)


def makeParser():
    """
    Make an argument parser.

    @return: An C{argparse} argument parser.
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Print SAM/BAM file reference sequence and unmapped read counts.",
    )

    parser.add_argument(
        "samFile", metavar="FILENAME", help="The name of a SAM/BAM alignment file."
    )

    parser.add_argument(
        "--sortBy",
        choices=("count", "coverage", "name"),
        default="count",
        help=(
            "Use 'count' to sort the output by decreasing read count (i.e., "
            "the reference with the highest number of matching reads is "
            "printed first), or 'coverage' to sort by decreasing genome "
            "coverage, or 'name' to sort by reference name."
        ),
    )

    parser.add_argument(
        "--excludeZeroes",
        action="store_true",
        help="Do not print references unmatched by any read.",
    )

    parser.add_argument(
        "--excludeIfNoAdditional",
        action="store_true",
        help=(
            "Do not print references that are mapped by no additional reads "
            "(i.e., reads other than those matching already-printed "
            "references). This is useful when --sortBy count is used, "
            "making it possible to not print references only matched by "
            "reads matching references that have already been printed."
        ),
    )

    parser.add_argument(
        "--topReferenceIdsFile",
        metavar="FILE",
        help=(
            "The file to write the (sorted) ids of the reads for the "
            "reference with the highest number of matching reads. Only "
            "valid when --sortBy count is used."
        ),
    )

    return parser


if __name__ == "__main__":
    main(makeParser().parse_args())



#!/usr/bin/env python

import argparse
import multiprocessing
import sys

from dark.aligners import mafft
from dark.fasta import FastaReads
from dark.reads import DNARead

MAFFT_DEFAULT_ARGS = "--globalpair --maxiterate 1000 --preservecase"
MAFFT_ALGORITHMS_URL = (
    "https://mafft.cbrc.jp/alignment/software/algorithms/algorithms.html"
)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Combine multiple sequences into a single sequence.",
    )

    parser.add_argument(
        "--fastaFiles",
        required=True,
        nargs="+",
        help=("The name of the FASTA files containing the sequences to be combined."),
    )

    parser.add_argument(
        "--threads",
        type=int,
        default=multiprocessing.cpu_count(),
        help=(
            "The number of threads to use when running the aligner (if "
            "--align is used and the alignment algorithm can make use of "
            "multiple threads)."
        ),
    )

    parser.add_argument(
        "--alignerOptions",
        default=MAFFT_DEFAULT_ARGS,
        help=(
            "Optional arguments to pass to the alignment algorithm. The "
            "default options are %r. See %s for some possible option "
            "combinations." % (MAFFT_DEFAULT_ARGS, MAFFT_ALGORITHMS_URL)
        ),
    )

    args = parser.parse_args()

    reads = []
    for fastaFile in args.fastaFiles:
        reads.extend(FastaReads(fastaFile))

    if len(reads) == 1:
        print("Cannot combine just one read. Exiting.", file=sys.stderr)
        sys.exit(1)

    alignedReads = mafft(reads, options=args.alignerOptions, threads=args.threads)

    combinedSequence = alignedReads.combineReads()

    print(DNARead("Combined", combinedSequence).toString(format_="fasta"))
